import datetime
import logging
import uuid
from typing import Dict, Literal, Optional

from oauthlib.oauth2.rfc6749.errors import InvalidClientError

from parsons import Table
from parsons.utilities import check_env
from parsons.utilities.oauth_api_connector import OAuth2APIConnector

logger = logging.getLogger(__name__)

ZOOM_URI = "https://api.zoom.us/v2/"
ZOOM_AUTH_CALLBACK = "https://zoom.us/oauth/token"

##########


class ZoomV1:
    def __init__(
        self,
        account_id: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
    ):
        """
        Instantiate the Zoom class.

        `Args:`
            account_id: str
                A valid Zoom account id. Not required if ``ZOOM_ACCOUNT_ID`` env
                variable set.
            client_id: str
                A valid Zoom client id. Not required if ``ZOOM_CLIENT_ID`` env
                variable set.
            client_secret: str
                A valid Zoom client secret. Not required if `ZOOM_CLIENT_SECRET` env
                variable set.
        """
        self.account_id = check_env.check("ZOOM_ACCOUNT_ID", account_id)
        self.client_id = check_env.check("ZOOM_CLIENT_ID", client_id)
        self.__client_secret = check_env.check("ZOOM_CLIENT_SECRET", client_secret)
        self.client = self.get_oauth_client()

    def get_oauth_client(self) -> OAuth2APIConnector:
        return OAuth2APIConnector(
            uri=ZOOM_URI,
            client_id=self.client_id,
            client_secret=self.__client_secret,
            token_url=ZOOM_AUTH_CALLBACK,
            auto_refresh_url=ZOOM_AUTH_CALLBACK,
            grant_type="account_credentials",
            authorization_kwargs={"account_id": self.account_id},
        )

    def _get_request(
        self,
        endpoint: str,
        data_key: Optional[str],
        params: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> Table:
        """
        TODO: Consider increasing default page size.

        `Args`:
            endpoint: str
                API endpoint to send GET request
            data_key: str
                Unique value to use to parse through nested data
                (akin to a primary key in response JSON)
            params: dict
                Additional request parameters, defaults to None

        `Returns`:
            Parsons Table of API responses
        """

        logger.warning("This version of the Zoom connector uses a deprecated pagination method.")
        logger.info("Consider switching to V2!")
        logger.info(
            "See docs for more information: https://move-coop.github.io/parsons/html/latest/zoom.html"
        )

        r = self.client.get_request(endpoint, params=params, **kwargs)
        self.client.data_key = data_key
        data = self.client.data_parse(r)

        if not params:
            params = {}

        # Return a dict or table if only one item.
        if "page_number" not in r.keys():
            if isinstance(data, dict):
                return data
            if isinstance(data, list):
                return Table(data)

        # Else iterate through the pages and return a Table
        else:
            while r["page_number"] < r["page_count"]:
                params["page_number"] = int(r["page_number"]) + 1
                r = self.client.get_request(endpoint, params=params, **kwargs)
                data.extend(self.client.data_parse(r))
            return Table(data)

    def __handle_nested_json(self, table: Table, column: str, version: int = 1) -> Table:
        """
        This function unpacks JSON values from Zoom's API, which are often
        objects nested in lists

        `Args`:
            table: parsons.Table
                Parsons Table of Zoom API responses

            column: str
                Column name of nested JSON

        `Returns`:
            Parsons Table
        """
        if version == 2:
            if column in table.columns:
                return Table(table.unpack_list(column=column))
            else:
                return table

        return Table(table.unpack_list(column=column)).unpack_dict(
            column=f"{column}_0", prepend_value=f"{column}_"
        )

    def __process_poll_results(self, tbl: Table) -> Table:
        """
        Unpacks nested poll results values from the Zoom reports endpoint

        `Args`:
            tbl: parsons.Table
                Table of poll results derived from Zoom API request

        `Returns`:
            Parsons Table
        """
        if tbl.num_rows == 0:
            return tbl

        # Add surrogate key
        tbl.add_column("poll_taker_id", lambda _: str(uuid.uuid4()))

        # Unpack values
        tbl = tbl.unpack_nested_columns_as_rows(
            "question_details", key="poll_taker_id", expand_original=True
        )

        # Remove extraneous columns
        tbl.remove_column("poll_taker_id")
        tbl.remove_column("question_details")

        # Unpack question values
        tbl = tbl.unpack_dict("question_details_value", include_original=True, prepend=False)

        # Remove column from API response
        tbl.remove_column("question_details_value")
        tbl.remove_column("uid")

        return tbl

    def get_users(self, status: str = "active", role_id: Optional[str] = None) -> Table:
        """
        Get users.

        `Args:`
            status: str
                Filter by the user status. Must be one of following: ``active``,
                ``inactive``, or ``pending``.
            role_id: str
                Filter by the user role.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        if status not in ["active", "inactive", "pending"]:
            raise ValueError("Invalid status type provided.")

        params = {"status": status, "role_id": role_id}

        tbl = self._get_request(endpoint="users", data_key="users", params=params)
        logger.info(f"Retrieved {tbl.num_rows} users.")
        return tbl

    def get_meetings(
        self,
        user_id: str,
        meeting_type: Literal[
            "scheduled", "live", "upcoming", "upcoming_meetings", "previous_meetings"
        ] = "scheduled",
        from_date: Optional[datetime.date] = None,
        to_date: Optional[datetime.date] = None,
    ) -> Table:
        """
        Get meetings scheduled by a user.

        `Args:`
            user_id: str
                A user id or email address of the meeting host.
            meeting_type: str

                .. list-table::
                    :widths: 25 50
                    :header-rows: 1

                    * - Type
                      - Notes
                    * - ``scheduled``
                      - This includes all valid past meetings, live meetings and upcoming
                        scheduled meetings. It is the equivalent to the combined list of
                        "Previous Meetings" and "Upcoming Meetings" displayed in the user's
                        Meetings page.
                    * - ``live``
                      - All the ongoing meetings.
                    * - ``upcoming``
                      - All upcoming meetings including live meetings.
            from_date: datetime.date or None
                Optional start date for the range of meetings to retrieve.
            to_date: datetime.date or None
                Optional end date for the range of meetings to retrieve.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        params: Dict[str, str] = {"type": meeting_type}
        if from_date:
            params["from"] = from_date.isoformat()
        if to_date:
            params["to"] = to_date.isoformat()

        tbl = self._get_request(f"users/{user_id}/meetings", "meetings", params=params)
        logger.info(f"Retrieved {tbl.num_rows} meetings.")
        return tbl

    def get_past_meeting(self, meeting_uuid: str) -> Table:
        """
        Get metadata regarding a past meeting.

        `Args:`
            meeting_id: int
                The meeting id
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = self._get_request(f"past_meetings/{meeting_uuid}", None)
        logger.info(f"Retrieved meeting {meeting_uuid}.")
        return tbl

    def get_past_meeting_participants(self, meeting_id: int) -> Table:
        """
        Get past meeting participants.

        `Args:`
            meeting_id: int
                The meeting id
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = self._get_request(f"report/meetings/{meeting_id}/participants", "participants")
        logger.info(f"Retrieved {tbl.num_rows} participants.")
        return tbl

    def get_meeting_registrants(self, meeting_id: int) -> Table:
        """
        Get meeting registrants.

        `Args:`
            meeting_id: int
                The meeting id
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = self._get_request(f"meetings/{meeting_id}/registrants", "registrants")
        logger.info(f"Retrieved {tbl.num_rows} registrants.")
        return tbl

    def get_user_webinars(self, user_id: str) -> Table:
        """
        Get meeting registrants.

        `Args:`
            user_id: str
                The user id
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = self._get_request(f"users/{user_id}/webinars", "webinars")
        logger.info(f"Retrieved {tbl.num_rows} webinars.")
        return tbl

    def get_past_webinar_report(self, webinar_id: str) -> Optional[Dict]:
        """
        Get past meeting participants

        `Args:`
            webinar_id: str
                The webinar id
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        dic = self._get_request(endpoint=f"report/webinars/{webinar_id}", data_key=None)
        if dic:
            logger.info(f"Retrieved webinar_report for webinar: {webinar_id}.")
        return dic

    def get_past_webinar_participants(self, webinar_id: str) -> Table:
        """
        Get past meeting participants

        `Args:`
            webinar_id: str
                The webinar id
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = self._get_request(f"report/webinars/{webinar_id}/participants", "participants")
        logger.info(f"Retrieved {tbl.num_rows} webinar participants.")
        return tbl

    def get_webinar_registrants(self, webinar_id: str) -> Table:
        """
        Get past meeting participants

        `Args:`
            webinar_id: str
                The webinar id
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = self._get_request(f"webinars/{webinar_id}/registrants", "registrants")
        logger.info(f"Retrieved {tbl.num_rows} webinar registrants.")
        return tbl

    def get_meeting_poll_metadata(self, meeting_id: int, poll_id: int, version: int = 1) -> Table:
        """
        Get metadata about a specific poll for a given meeting ID

        Required scopes: `meeting:read`

        `Args`:
            meeting_id: int
                Unique identifier for Zoom meeting
            poll_id: int
                Unique identifier for poll

        `Returns`:
            Parsons Table of all polling responses
        """

        endpoint = f"meetings/{meeting_id}/polls/{poll_id}"
        tbl = self._get_request(endpoint=endpoint, data_key="questions")

        if tbl.num_rows == 0:
            logger.debug(f"No poll data returned for poll ID {poll_id}")
            return tbl

        logger.info(
            f"Retrieved {tbl.num_rows} rows of metadata [meeting={meeting_id} poll={poll_id}]"
        )

        if "prompts" in tbl.columns:
            logger.info(f"Unnesting columns 'prompts' from existing table columns: {tbl.columns}")
            return self.__handle_nested_json(table=tbl, column="prompts", version=version)
        else:
            return tbl

    def get_meeting_all_polls_metadata(self, meeting_id: int, version: int = 1) -> Table:
        """
        Get metadata for all polls for a given meeting ID

        Required scopes: `meeting:read`

        `Args`:
            meeting_id: int
                Unique identifier for Zoom meeting

        `Returns`:
            Parsons Table of all polling responses
        """

        endpoint = f"meetings/{meeting_id}/polls"
        tbl = self._get_request(endpoint=endpoint, data_key="polls")

        if tbl.num_rows == 0:
            logger.debug(f"No poll data returned for meeting ID {meeting_id}")
            return tbl

        logger.info(f"Retrieved {tbl.num_rows} polls for meeting ID {meeting_id}")

        return self.__handle_nested_json(table=tbl, column="questions", version=version)

    def get_past_meeting_poll_metadata(self, meeting_id: int, version: int = 1) -> Table:
        """
        List poll metadata of a past meeting.

        Required scopes: `meeting:read`

        `Args`:
            meeting_id: int
                The meeting's ID or universally unique ID (UUID).

        `Returns`:
            Parsons Table of poll results
        """

        endpoint = f"past_meetings/{meeting_id}/polls"
        tbl = self._get_request(endpoint=endpoint, data_key="questions")

        if tbl.num_rows == 0:
            logger.debug(f"No poll data returned for meeting ID {meeting_id}")
            return tbl

        logger.info(f"Retrieved {tbl.num_rows} polls for meeting ID {meeting_id}")
        logger.info(
            f"Unnesting columns 'question_details' from existing table columns: {tbl.columns}"
        )

        return self.__handle_nested_json(table=tbl, column="question_details", version=version)

    def get_webinar_poll_metadata(self, webinar_id: str, poll_id: int, version: int = 1) -> Table:
        """
        Get metadata for a specific poll for a given webinar ID

        Required scopes: `webinar:read`

        `Args`:
            webinar_id: str
                Unique identifier for Zoom webinar
            poll_id: int
                Unique identifier for poll

        `Returns`:
            Parsons Table of all polling responses
        """

        endpoint = f"webinars/{webinar_id}/polls/{poll_id}"
        tbl = self._get_request(endpoint=endpoint, data_key="questions")

        if tbl.num_rows == 0:
            logger.debug(f"No poll data returned for poll ID {poll_id}")
            return tbl

        logger.info(
            f"Retrieved {tbl.num_rows} rows of metadata [meeting={webinar_id} poll={poll_id}]"
        )

        return self.__handle_nested_json(table=tbl, column="prompts", version=version)

    def get_webinar_all_polls_metadata(self, webinar_id: str, version: int = 1) -> Table:
        """
        Get metadata for all polls for a given webinar ID

        Required scopes: `webinar:read`

        `Args`:
            webinar_id: str
                Unique identifier for Zoom webinar

        `Returns`:
            Parsons Table of all polling responses
        """

        endpoint = f"webinars/{webinar_id}/polls"
        tbl = self._get_request(endpoint=endpoint, data_key="polls")

        if tbl.num_rows == 0:
            logger.debug(f"No poll data returned for webinar ID {webinar_id}")
            return tbl

        logger.info(f"Retrieved {tbl.num_rows} polls for meeting ID {webinar_id}")

        return self.__handle_nested_json(table=tbl, column="questions", version=version)

    def get_past_webinar_poll_metadata(self, webinar_id: str, version: int = 1) -> Table:
        """
        Retrieves the metadata for Webinar Polls of a specific Webinar

        Required scopes: `webinar:read`

        `Args`:
            webinar_id: str
                The webinar's ID or universally unique ID (UUID).

        `Returns`:
            Parsons Table of all polling responses
        """

        endpoint = f"past_webinars/{webinar_id}/polls"
        tbl = self._get_request(endpoint=endpoint, data_key="questions")

        if tbl.num_rows == 0:
            logger.debug(f"No poll data returned for webinar ID {webinar_id}")
            return tbl

        logger.info(f"Retrieved {tbl.num_rows} polls for meeting ID {webinar_id}")

        return self.__handle_nested_json(table=tbl, column="question_details", version=version)

    def get_meeting_poll_results(self, meeting_id: int) -> Table:
        """
        Get a report of poll results for a past meeting

        Required scopes: `report:read:admin`
        """

        endpoint = f"report/meetings/{meeting_id}/polls"
        tbl = self._get_request(endpoint=endpoint, data_key="questions")

        if tbl.num_rows == 0:
            logger.debug(f"No poll data returned for meeting ID {meeting_id}")
            return tbl

        logger.info(f"Retrieved {tbl.num_rows} reults for meeting ID {meeting_id}")

        return self.__process_poll_results(tbl=tbl)

    def get_webinar_poll_results(self, webinar_id: str) -> Table:
        """
        Get a report of poll results for a past webinar

        Required scopes: `report:read:admin`
        """

        endpoint = f"report/webinars/{webinar_id}/polls"
        tbl = self._get_request(endpoint=endpoint, data_key="questions")

        if tbl.num_rows == 0:
            logger.debug(f"No poll data returned for webinar ID {webinar_id}")
            return tbl

        logger.info(f"Retrieved {tbl.num_rows} reults for webinar ID {webinar_id}")

        return self.__process_poll_results(tbl=tbl)


class ZoomV2(ZoomV1):
    """
    Version 2 implementation of a Parsons connector. Designed to involve minimal
    transformation logic and clearer naming conventions.

    Inherits the following methods from version 1:
    - get_users
    - get_meetings
    - get_past_meeting
    - get_meeting_registrants
    - get_past_webinar_report
    - get_webinar_registrants

    Overwrites the following methods from version 1:
    - get_past_meeting_participants
    - get_past_webinar_participants

    Renames/refactors the following methods (and raises error if original function called):
    - get_user_webinars
    - get_meeting_poll_metadata
    - get_meeting_all_polls_metadata
    - get_past_meeting_poll_metadata
    - get_webinar_poll_metadata
    - get_webinar_all_polls_metadata
    - get_past_webinar_poll_metadata
    - get_meeting_poll_results
    - get_webinar_poll_results

    Args:
        ZoomV1 (cls): version 1 Zoom connector class
    """

    def __init__(
        self,
        account_id: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
    ):
        super().__init__(account_id, client_id, client_secret)

    def _get_request(
        self,
        endpoint: str,
        data_key: Optional[str],
        params: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> Table:
        """
        `Args`:
            endpoint: str
                API endpoint to send GET request
            data_key: str
                Unique value to use to parse through nested data
                (akin to a primary key in response JSON)
            params: dict
                Additional request parameters, defaults to None

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        if params is None:
            params = {}
        if not params:
            params = {"page_size": 300}

        self.client.data_key = data_key
        next_page_token = ""
        has_more_pages = True
        data = []

        while has_more_pages:
            if next_page_token:
                params["next_page_token"] = next_page_token

            try:
                r = self.client.get_request(endpoint, params=params, **kwargs)
            except InvalidClientError:
                self.client = self.get_oauth_client()
                r = self.client.get_request(endpoint, params=params, **kwargs)
            parsed_resp = self.client.data_parse(r)
            if isinstance(parsed_resp, dict):
                parsed_resp = [parsed_resp]
            data.extend(parsed_resp)

            next_page_token = r.get("next_page_token")
            has_more_pages = bool(next_page_token)

        return Table(data)

    def get_webinars(self, user_id: int) -> Table:
        """
        Get webinars scheduled by or on behalf of a webinar host.

        `Args:`
            user_id: str
                The user id
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = self._get_request(f"users/{user_id}/webinars", "webinars")
        logger.info(f"Retrieved {tbl.num_rows} webinars.")
        return tbl

    def get_webinar_occurrences(self, webinar_id: int) -> Table:
        """
        Get webinar occurrences for a given webinar ID.

        `Args:`
            webinar_id: int
                The webinar id
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        tbl = self._get_request(f"webinars/{webinar_id}/", "occurrences")
        logger.info(f"Retrieved {tbl.num_rows} webinar occurrences.")
        return tbl

    def get_user_webinars(self, user_id: str) -> AttributeError:
        return AttributeError(
            "Method get_user_webinars has been deprecated in favor of get_webinars"
        )

    def get_past_meeting_participants(self, meeting_id: int) -> Table:
        """
        Get past meeting participants.

        `Args:`
            meeting_id: int
                The meeting id
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = self._get_request(f"past_meetings/{meeting_id}/participants", "participants")
        logger.info(f"Retrieved {tbl.num_rows} participants.")
        return tbl

    def get_past_webinar_participants(self, webinar_id: int) -> Table:
        """
        Get past webinar participants.

        `Args:`
            webinar_id: int
                The webinar id
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = self._get_request(f"past_webinars/{webinar_id}/participants", "participants")
        logger.info(f"Retrieved {tbl.num_rows} participants.")
        return tbl

    def get_meeting_poll(self, meeting_id: int, poll_id: str) -> Table:
        """
        Get information about a single poll for a given meeting ID.
        The returned data is identical to get_meeting_polls.

        `Args`:
            meeting_id: int
                Unique identifier for Zoom meeting
            poll_id: str
                Unique identifier for Zoom poll

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        endpoint = f"meetings/{meeting_id}/polls/{poll_id}"
        tbl = self._get_request(endpoint=endpoint, data_key=None)
        logger.info(f"Retrieved {tbl.num_rows} for [poll {poll_id}, meeting {meeting_id}]")
        return tbl

    def get_meeting_poll_metadata(self, meeting_id, poll_id, version=1):
        raise AttributeError(
            "Method get_meeting_poll_metadata is deprecated in favor of get_meeting_poll"
        )

    def get_meeting_polls(self, meeting_id: int) -> Table:
        """
        Get information about all polls for a given meeting ID.
        The returned data is identical to get_meeting_poll but for
        all polls in the meeting.

        `Args`:
            meeting_id: int
                Unique identifier for Zoom meeting

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        endpoint = f"meetings/{meeting_id}/polls"
        tbl = self._get_request(endpoint=endpoint, data_key="polls")
        logger.info(f"Retrieved {tbl.num_rows} meeting polls for meeting {meeting_id}")
        return tbl

    def get_meeting_all_polls_metadata(self, meeting_id, version=1):
        raise AttributeError(
            "Method get_meeting_all_polls_metadata is deprecated in favor of get_meeting_polls"
        )

    def get_past_meeting_poll_results(self, meeting_id: int) -> Table:
        """
        Get results for all polls for a given past meeting ID

        `Args`:
            meeting_id: int
                Unique identifier for Zoom meeting

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        endpoint = f"past_meetings/{meeting_id}/polls"
        tbl = self._get_request(endpoint=endpoint, data_key=None)
        logger.info(f"Retrieved {tbl.num_rows} meeting poll results")
        return tbl

    def get_past_meeting_poll_metadata(self, meeting_id, version=1):
        raise AttributeError(
            "Method get_past_meeting_poll_metadata is deprecated in favor of get_past_meeting_poll_results"
        )

    def get_webinar_poll(self, webinar_id: int, poll_id: str) -> Table:
        """
        Get information about a single poll for a given webinar ID.
        The returned data is identical to get_webinar_polls.

        `Args`:
            webinar_id: int
                Unique identifier for Zoom webinar
            poll_id: str
                Unique identifier for Zoom poll

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        endpoint = f"webinars/{webinar_id}/polls/{poll_id}"
        tbl = self._get_request(endpoint=endpoint, data_key=None)
        logger.info(f"Retrieved {tbl.num_rows} for [poll {poll_id}, webinar {webinar_id}]")
        return tbl

    def get_webinar_poll_metadata(self, webinar_id, poll_id, version=1):
        raise AttributeError(
            "Method get_webinar_poll_metadata is deprecated in favor of get_webinar_poll"
        )

    def get_webinar_polls(self, webinar_id: int) -> Table:
        """
        Get information for all polls for a given webinar ID
        The returned data is identical to get_webinar_poll but includes
        all polls in the webinar

        `Args`:
            webinar_id: str
                Unique identifier for Zoom webinar

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        endpoint = f"webinars/{webinar_id}/polls"
        tbl = self._get_request(endpoint=endpoint, data_key="polls")
        logger.info(f"Retrieved {tbl.num_rows} polls for webinar ID {webinar_id}")
        return tbl

    def get_webinar_all_polls_metadata(self, webinar_id, version=1):
        raise AttributeError(
            "Method get_webinar_all_polls_metadata is deprecated in favor of get_webinar_polls"
        )

    def get_past_webinar_poll_results(self, webinar_id: int) -> Table:
        """
        Get results for all polls for a given past webinar ID

        `Args`:
            webinar_id: str
                Unique identifier for Zoom webinar

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        endpoint = f"past_webinars/{webinar_id}/polls"
        tbl = self._get_request(endpoint=endpoint, data_key=None)
        logger.info(f"Retrieved {tbl.num_rows} poll results for webinar ID {webinar_id}")
        return tbl

    def get_past_webinar_poll_metadata(self, webinar_id, version=1):
        raise AttributeError(
            "Method get_past_webinar_poll_metadata is deprecated in favor of get_past_webinar_poll_results"
        )

    def get_meeting_poll_reports(self, meeting_id: int) -> Table:
        """
        Get polls reports for a given past meeting ID

        `Args`:
            meeting_id: str
                Unique identifier for Zoom meeting

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        endpoint = f"report/meetings/{meeting_id}/polls"
        tbl = self._get_request(endpoint=endpoint, data_key=None)
        logger.info(f"Retrieved {tbl.num_rows} poll reports for meeting ID {meeting_id}")
        return tbl

    def get_meeting_poll_results(self, meeting_id):
        raise AttributeError(
            "Method get_meeting_poll_results is deprecated in favor of get_meeting_poll_reports"
        )

    def get_webinar_poll_reports(self, webinar_id: int) -> Table:
        """
        Get results for all polls for a given past webinar ID

        `Args`:
            webinar_id: str
                Unique identifier for Zoom webinar

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        endpoint = f"report/webinars/{webinar_id}/polls"
        tbl = self._get_request(endpoint=endpoint, data_key=None)
        logger.info(f"Retrieved {tbl.num_rows} poll reports for webinar ID {webinar_id}")
        return tbl

    def get_webinar_poll_results(self, webinar_id):
        raise AttributeError(
            "Method get_webinar_poll_results is deprecated in favor of get_webinar_poll_reports"
        )


class Zoom:
    def __new__(
        cls,
        account_id: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        parsons_version: str = "v1",
    ) -> ZoomV1:
        """
        Create and return Zoom instance base on chosen version (1 or 2)

        API Documentation: https://developers.zoom.us/docs/api/

        Args:
            account_id: str
                A valid Zoom account id. Not required if ``ZOOM_ACCOUNT_ID`` env
                variable set.
            client_id: str
                A valid Zoom client id. Not required if ``ZOOM_CLIENT_ID`` env
                variable set.
            client_secret: str
                A valid Zoom client secret. Not required if `ZOOM_CLIENT_SECRET` env
                variable set.
            parsons_version (str, optional): Parsons version of the Zoom connector. Defaults to v1.
        """
        parsons_version = check_env.check("ZOOM_PARSONS_VERSION", parsons_version)
        if parsons_version == "v1":
            logger.info("Consider upgrading to version 2 of the Zoom connector!")
            logger.info(
                "See docs for more information: https://move-coop.github.io/parsons/html/latest/zoom.html"
            )
            return ZoomV1(account_id=account_id, client_id=client_id, client_secret=client_secret)
        if parsons_version == "v2":
            return ZoomV2(account_id=account_id, client_id=client_id, client_secret=client_secret)
        raise ValueError(f"{parsons_version} not supported")



from datetime import datetime

from dateutil.parser import parse


class Flags:
    """A class to access the Flags PDI API endpoint."""

    def __init__(self):
        super().__init__()
        flags_endpoint = "/flags"
        self.url_flags = self.base_url + flags_endpoint

    def get_flags(self, start_date, end_date, limit=None):
        """Get a list of flags.

        `Args:`
            start_date: str
                A start date formatted like yyyy-MM-dd.
            end_date: str
                An end date formatted like yyyy-MM-dd.
            limit: int
                Specify limit to return.

        `Returns:`
            parsons.Table
                A Parsons table of all the data.
        """
        try:
            start_date = parse(start_date).date().isoformat()
            end_date = parse(end_date).date().isoformat()
        except ValueError as e:
            raise ValueError("Invalid date format.") from e

        params = {
            "startDate": start_date,
            "endDate": end_date,
        }

        return self._request(self.url_flags, args=params, limit=limit)

    def create_flags(self, flag_list: list):
        """
        Save a list of flags, each flag must look like the dictionary below
        [
            {
                "pdiId": "string",
                "flagEntryDate": An end date formatted like yyyy-MM-dd.,
                "acquisitionTypeId": "string",
                "flagId": "string",
                "questionId": "string",
                "contactId": "string"
            }
        ]
        """
        if "pdiId" not in list(flag_list[0].keys()):
            raise ValueError("missing required key")
            return {}
        for flag in flag_list:
            try:
                flag["flagEntryDate"] = str(
                    datetime.strptime(flag["flagEntryDate"], "%Y-%m-%d").isoformat()
                )
            except ValueError as e:
                raise ValueError("Invalid date format.") from e
        print(flag_list)
        return self._request(self.url_flags, post_data=flag_list, req_type="POST")

    def delete_flag(self, id: str):
        """
        Delete a Flag by id.
        `Args:`
            id: str
                The Flag id
        """
        return self._request(f"self.url_flags/{id}", req_type="DELETE")



import unittest

import requests_mock

from parsons import Mailchimp
from test.test_mailchimp import expected_json

API_KEY = "mykey-us00"


class TestMailchimp(unittest.TestCase):
    def setUp(self):
        self.mc = Mailchimp(API_KEY)

    @requests_mock.Mocker()
    def test_get_campaigns(self, m):
        # Test that campaigns are returned correctly.
        m.get(self.mc.uri + "campaigns", json=expected_json.test_campaigns)
        tbl = self.mc.get_campaigns()

        self.assertEqual(tbl.num_rows, 2)

    @requests_mock.Mocker()
    def test_get_lists(self, m):
        # Test that lists are returned correctly.
        m.get(self.mc.uri + "lists", json=expected_json.test_lists)
        tbl = self.mc.get_lists()

        self.assertEqual(tbl.num_rows, 2)

    @requests_mock.Mocker()
    def test_get_members(self, m):
        # Test that list members are returned correctly.
        m.get(self.mc.uri + "lists/zyx/members", json=expected_json.test_members)
        tbl = self.mc.get_members(list_id="zyx")

        self.assertEqual(tbl.num_rows, 2)

    @requests_mock.Mocker()
    def test_get_unsubscribes(self, m):
        # Test that campaign unsubscribes are returned correctly.
        m.get(
            self.mc.uri + "reports/abc/unsubscribed",
            json=expected_json.test_unsubscribes,
        )
        tbl = self.mc.get_unsubscribes(campaign_id="abc")

        self.assertEqual(tbl.num_rows, 1)



import logging

import petl

logger = logging.getLogger(__name__)


class ETL(object):
    def __init__(self):
        pass

    def head(self, n=5):
        """
        Return the first n rows of the table

        `Args:`
            n: int
                The number of rows to return. Defaults to 5.
        `Returns:`
            `Parsons Table`
        """

        self.table = petl.head(self.table, n)

        return self

    def tail(self, n=5):
        """
        Return the last n rows of the table. Defaults to 5.

        `Args:`
            n: int
                The number of rows to return

        `Returns:`
            `Parsons Table`
        """

        self.table = petl.tail(self.table, n)

        return self

    def add_column(self, column, value=None, index=None, if_exists="fail"):
        """
        Add a column to your table

        `Args:`
            column: str
                Name of column to add
            value:
                A fixed or calculated value
            index: int
                The position of the new column in the table
            if_exists: str (options: 'fail', 'replace')
                If set `replace`, this function will call `fill_column`
                if the column already exists, rather than raising a `ValueError`
        `Returns:`
            `Parsons Table` and also updates self
        """

        if column in self.columns:
            if if_exists == "replace":
                self.fill_column(column, value)
                return self
            else:
                raise ValueError(f"Column {column} already exists")

        self.table = self.table.addfield(column, value, index)

        return self

    def remove_column(self, *columns):
        r"""
        Remove a column from your table

        `Args:`
            *columns: str
                Column names
        `Returns:`
            `Parsons Table` and also updates self
        """

        self.table = petl.cutout(self.table, *columns)

        return self

    def rename_column(self, column_name, new_column_name):
        """
        Rename a column

        `Args:`
            column_name: str
                The current column name
            new_column_name: str
                The new column name
        `Returns:`
            `Parsons Table` and also updates self
        """

        if new_column_name in self.columns:
            raise ValueError(f"Column {new_column_name} already exists")

        self.table = petl.rename(self.table, column_name, new_column_name)

        return self

    def rename_columns(self, column_map):
        """
        Rename multiple columns

        `Args:`
            column_map: dict
                A dictionary of columns and new names.
                The key is the old name and the value is the new name.

                Example dictionary:
                {'old_name': 'new_name',
                'old_name2': 'new_name2'}

        `Returns:`
            `Parsons Table` and also updates self
        """

        # Check if old column name exists and new column name does not exist
        for old_name, new_name in column_map.items():
            if old_name not in self.table.columns():
                raise KeyError(f"Column name {old_name} does not exist")
            if new_name in self.table.columns():
                raise ValueError(f"Column name {new_name} already exists")

        # Uses the underlying petl method
        self.table = petl.rename(self.table, column_map)

        return self

    def fill_column(self, column_name, fill_value):
        """
        Fill a column in a table

        `Args:`
            column_name: str
                The column to fill
            fill_value:
                A fixed or calculated value
        `Returns:`
            `Parsons Table` and also updates self
        """

        if callable(fill_value):
            self.table = petl.convert(
                self.table, column_name, lambda _, r: fill_value(r), pass_row=True
            )
        else:
            self.table = petl.update(self.table, column_name, fill_value)

        return self

    def fillna_column(self, column_name, fill_value):
        """
        Fill None values in a column in a table

        `Args:`
            column_name: str
                The column to fill
            fill_value:
                A fixed or calculated value
        `Returns:`
            `Parsons Table` and also updates self
        """

        if callable(fill_value):
            self.table = petl.convert(
                self.table,
                column_name,
                lambda _, r: fill_value(r),
                where=lambda r: r[column_name] is None,
                pass_row=True,
            )
        else:
            self.table = petl.update(
                self.table,
                column_name,
                fill_value,
                where=lambda r: r[column_name] is None,
            )

        return self

    def move_column(self, column, index):
        """
        Move a column

        `Args:`
            column: str
                The column name to move
            index:
                The new index for the column
        `Returns:`
            `Parsons Table` and also updates existing object.
        """

        self.table = petl.movefield(self.table, column, index)

        return self

    def convert_column(self, *column, **kwargs):
        """
        Transform values under one or more fields via arbitrary functions, method
        invocations or dictionary translations. This leverages the petl ``convert()``
        method. Example usage can be found `here <https://petl.readthedocs.io/latest/transform.html#petl.transform.conversions.convert>`_.

        `Args:`
            *column: str
                A single column or multiple columns passed as a list
            **kwargs: str, method or variable
                The update function, method, or variable to process the update
        `Returns:`
            `Parsons Table` and also updates self
        """

        self.table = petl.convert(self.table, *column, **kwargs)

        return self

    def get_column_max_width(self, column):
        """
        Return the maximum width of the column.

        `Args:`
            column: str
                The column name.
        `Returns:`
            int
        """

        max_width = 0

        for v in petl.values(self.table, column):
            if len(str(v).encode("utf-8")) > max_width:
                max_width = len(str(v).encode("utf-8"))

        return max_width

    def convert_columns_to_str(self):
        """
        Convenience function to convert all non-string or mixed columns in a
        Parsons table to string (e.g. for comparison)

        `Returns:`
            `Parsons Table` and also updates self
        """

        # If we don't have any rows, don't bother trying to convert things
        if self.num_rows == 0:
            return self

        cols = self.get_columns_type_stats()

        def str_or_empty(x):
            if x is None:
                return ""
            return str(x)

        for col in cols:
            # If there's more than one type (or no types), convert to str
            # Also if there is one type and it's not str, convert to str
            if len(col["type"]) != 1 or col["type"][0] != "str":
                self.convert_column(col["name"], str_or_empty)

        return self

    def coalesce_columns(self, dest_column, source_columns, remove_source_columns=True):
        """
        Coalesces values from one or more source columns into a destination column, by selecting
        the first non-empty value. If the destination column doesn't exist, it will be added.

        `Args:`
            dest_column: str
                Name of destination column
            source_columns: list
                List of source column names
            remove_source_columns: bool
                Whether to remove the source columns after the coalesce. If the destination
                column is also one of the source columns, it will not be removed.
        `Returns:`
            `Parsons Table` and also updates self
        """

        if dest_column in self.columns:

            def convert_fn(value, row):
                for source_col in source_columns:
                    if row.get(source_col):
                        return row[source_col]

            logger.debug(f"Coalescing {source_columns} into {dest_column}")
            self.convert_column(dest_column, convert_fn, pass_row=True)

        else:

            def add_fn(row):
                for source_col in source_columns:
                    if row.get(source_col):
                        return row[source_col]

            logger.debug(f"Creating new column {dest_column} from {source_columns}")
            self.add_column(dest_column, add_fn)

        if remove_source_columns:
            for source_col in source_columns:
                if source_col != dest_column:
                    self.remove_column(source_col)

        return self

    def map_columns(self, column_map, exact_match=True):
        """
        Standardizes column names based on multiple possible values. This method
        is helpful when your input table might have multiple and unknown column
        names.

        `Args:`
            column_map: dict
                A dictionary of columns and possible values that map to it
            exact_match: boolean
                If ``True`` will only map if an exact match. If ``False`` will
                ignore case, spaces and underscores.
        `Returns:`
            `Parsons Table` and also updates self

        .. code-block:: python

            tbl = [{'fn': 'Jane'},
                   {'lastname': 'Doe'},
                   {'dob': '1980-01-01'}]
            column_map = {'first_name': ['fn', 'first', 'firstname'],
                          'last_name': ['ln', 'last', 'lastname'],
                          'date_of_birth': ['dob', 'birthday']}

            tbl.map_columns(column_map)
            print (tbl)
            >> {{'first_name': 'Jane', 'last_name': 'Doe', 'date_of_birth': '1908-01-01'}}
        """

        for col in self.columns:
            if not exact_match:
                cleaned_col = col.lower().replace("_", "").replace(" ", "")
            else:
                cleaned_col = col

            for k, v in column_map.items():
                for i in v:
                    if cleaned_col == i:
                        self.rename_column(col, k)

        return self

    def map_and_coalesce_columns(self, column_map):
        """
        Coalesces columns based on multiple possible values. The columns in the map
        do not need to be in your table, so you can create a map with all possibilities.
        The coalesce will occur in the order that the columns are listed, unless the
        destination column name already exists in the table, in which case that
        value will be preferenced. This method is helpful when your input table might
        have multiple and unknown column names.
        `Args:`
            column_map: dict
                A dictionary of columns and possible values that map to it

        `Returns:`
            `Parsons Table` and also updates self

        .. code-block:: python

            tbl = [{'first': None},
                   {'fn': 'Jane'},
                   {'lastname': 'Doe'},
                   {'dob': '1980-01-01'}]

            column_map = {'first_name': ['fn', 'first', 'firstname'],
                          'last_name': ['ln', 'last', 'lastname'],
                          'date_of_birth': ['dob', 'birthday']}

            tbl.map_and_coalesce_columns(column_map)

            print (tbl)
            >> {{'first_name': 'Jane', 'last_name': 'Doe', 'date_of_birth': '1908-01-01'}}
        """

        for key, value in column_map.items():
            coalesce_list = value
            # if the column in the mapping dict isn't actually in the table,
            # remove it from the list of columns to coalesce
            for item in coalesce_list:
                if item not in self.columns:
                    coalesce_list.remove(item)
            # if the key from the mapping dict already exists in the table,
            # rename it so it can be coalesced with other possible columns
            if key in self.columns:
                self.rename_column(key, f"{key}_temp")
                coalesce_list.insert(0, f"{key}_temp")

            # coalesce columns
            self.coalesce_columns(key, coalesce_list, remove_source_columns=True)

        return self

    def get_column_types(self, column):
        """
        Return all of the Python types for values in a given column

        `Args:`
            column: str
                Name of the column to analyze
        `Returns:`
            list
                A list of Python types
        """

        return list(petl.typeset(self.table, column))

    def get_columns_type_stats(self):
        """
        Return descriptive stats for all columns

        `Returns:`
            list
                A list of dicts
        `Returns:`
            list
                A list of dicts, each containing a column 'name' and a 'type' list
        """

        return [{"name": col, "type": self.get_column_types(col)} for col in self.table.columns()]

    def convert_table(self, *args):
        r"""
        Transform all cells in a table via arbitrary functions, method invocations or dictionary
        translations. This method is useful for cleaning fields and data hygiene functions such
        as regex. This method leverages the petl ``convert()`` method. Example usage can be
        found `here` <https://petl.readthedocs.io/en/v0.24/transform.html#petl.convert>`_.

        `Args:`
            *args: str, method or variable
                The update function, method, or variable to process the update. Can also
        `Returns:`
            `Parsons Table` and also updates self
        """

        self.convert_column(self.columns, *args)

        return self

    def unpack_dict(
        self,
        column,
        keys=None,
        include_original=False,
        sample_size=5000,
        missing=None,
        prepend=True,
        prepend_value=None,
    ):
        """
        Unpack dictionary values from one column into separate columns

        `Args:`
            column: str
                The column name to unpack
            keys: list
                The dict keys in the column to unpack. If ``None`` will unpack
                all.
            include_original: boolean
                Retain original column after unpacking
            sample_size: int
                Number of rows to sample before determining columns
            missing: str
                If a value is missing, the value to fill it with
            prepend:
                Prepend the column name of the unpacked values. Useful for
                avoiding duplicate column names
            prepend_value:
                Value to prepend new columns if ``prepend=True``. If None, will
                set to column name.
        """

        if prepend:
            if prepend_value is None:
                prepend_value = column

            self.table = petl.convert(
                self.table, column, lambda v: self._prepend_dict(v, prepend_value)
            )

        self.table = petl.unpackdict(
            self.table,
            column,
            keys=keys,
            includeoriginal=include_original,
            samplesize=sample_size,
            missing=missing,
        )

        return self

    def unpack_list(
        self,
        column,
        include_original=False,
        missing=None,
        replace=False,
        max_columns=None,
    ):
        """
        Unpack list values from one column into separate columns. Numbers the
        columns.

        .. code-block:: python

          # Begin with a list in column
          json = [{'id': '5421',
                   'name': 'Jane Green',
                   'phones': ['512-699-3334', '512-222-5478']
                  }
                 ]

          tbl = Table(json)
          print (tbl)
          >>> {'id': '5421', 'name': 'Jane Green', 'phones': ['512-699-3334', '512-222-5478']}

          tbl.unpack_list('phones', replace=True)
          print (tbl)
          >>> {'id': '5421', 'name': 'Jane Green', 'phones_0': '512-699-3334', 'phones_1': '512-222-5478'}

        `Args:`
            column: str
                The column name to unpack
            include_original: boolean
                Retain original column after unpacking
            sample_size: int
                Number of rows to sample before determining columns
            missing: str
                If a value is missing, the value to fill it with
            replace: boolean
                Return new table or replace existing
            max_columns: int
                The maximum number of columns to unpack
        `Returns:`
            None
        """

        # Convert all column values to list to avoid unpack errors
        self.table = petl.convert(
            self.table, column, lambda v: [v] if not isinstance(v, list) else v
        )

        # Find the max number of values in list for all rows
        col_count = 0
        for row in self.cut(column):
            if len(row[column]) > col_count:
                col_count = len(row[column])

        # If max columns provided, set max columns
        if col_count > 0 and max_columns:
            col_count = max_columns

        # Create new column names "COL_01, COL_02"
        new_cols = []
        for i in range(col_count):
            new_cols.append(column + "_" + str(i))

        tbl = petl.unpack(
            self.table,
            column,
            new_cols,
            include_original=include_original,
            missing=missing,
        )

        if replace:
            self.table = tbl

        else:
            return tbl

    def unpack_nested_columns_as_rows(self, column, key="id", expand_original=False):
        """
        Unpack list or dict values from one column into separate rows.
        Not recommended for JSON columns (i.e. lists of dicts), but can handle columns
        with any mix of types. Makes use of PETL's `melt()` method.

        `Args:`
            column: str
                The column name to unpack
            key: str
                The column to use as a key when unpacking. Defaults to `id`
            expand_original: boolean or int
                If `True`: Add resulting unpacked rows (with all other columns) to original
                If `int`: Add to original unless the max added per key is above the given number
                If `False` (default): Return unpacked rows (with `key` column only) as standalone
                Removes packed list and dict rows from original either way.
        `Returns:`
            If `expand_original`, original table with packed rows replaced by unpacked rows
            Otherwise, standalone table with key column and unpacked values only
        """

        if isinstance(expand_original, int) and expand_original is not True:
            lengths = {len(row[column]) for row in self if isinstance(row[column], (dict, list))}
            max_len = sorted(lengths, reverse=True)[0]
            if max_len > expand_original:
                expand_original = False

        if expand_original:
            # Include all columns and filter out other non-dict types in table_list
            table = self
            table_list = table.select_rows(lambda row: isinstance(row[column], list))
        else:
            # Otherwise, include only key and column, but keep all non-dict types in table_list
            table = self.cut(key, column)
            table_list = table.select_rows(lambda row: not isinstance(row[column], dict))

        # All the columns other than column to ignore while melting
        ignore_cols = table.columns
        ignore_cols.remove(column)

        # Unpack lists as separate columns
        table_list.unpack_list(column, replace=True)

        # Rename the columns to retain only the number
        for col in table_list.columns:
            if f"{column}_" in col:
                table_list.rename_column(col, col.replace(f"{column}_", ""))

        # Filter dicts and unpack as separate columns
        table_dict = table.select_rows(lambda row: isinstance(row[column], dict))
        table_dict.unpack_dict(column, prepend=False)

        from parsons.etl.table import Table

        # Use melt to pivot both sets of columns into their own Tables and clean out None values
        melted_list = Table(petl.melt(table_list.table, ignore_cols))
        melted_dict = Table(petl.melt(table_dict.table, ignore_cols))

        melted_list.remove_null_rows("value")
        melted_dict.remove_null_rows("value")

        melted_list.rename_column("variable", column)
        melted_dict.rename_column("variable", column)

        # Combine the list and dict Tables
        melted_list.concat(melted_dict)

        import hashlib

        if expand_original:
            # Add unpacked rows to the original table (minus packed rows)
            orig = self.select_rows(lambda row: not isinstance(row[column], (dict, list)))
            orig.concat(melted_list)
            # Add unique id column by hashing all the other fields
            if "uid" not in self.columns:
                orig.add_column(
                    "uid",
                    lambda row: hashlib.md5(
                        str.encode("".join([str(x) for x in row])), usedforsecurity=False
                    ).hexdigest(),
                )
                orig.move_column("uid", 0)

            # Rename value column in case this is done again to this Table
            orig.rename_column("value", f"{column}_value")

            # Keep column next to column_value
            orig.move_column(column, -1)
            output = orig
        else:
            orig = self.remove_column(column)
            # Add unique id column by hashing all the other fields
            melted_list.add_column(
                "uid",
                lambda row: hashlib.md5(
                    str.encode("".join([str(x) for x in row])), usedforsecurity=False
                ).hexdigest(),
            )
            melted_list.move_column("uid", 0)
            output = melted_list

        self = orig
        return output

    def long_table(
        self,
        key,
        column,
        key_rename=None,
        retain_original=False,
        prepend=True,
        prepend_value=None,
    ):
        """
        Create a new long parsons table from a column, including the foreign
        key.

        .. code-block:: python

           # Begin with nested dicts in a column
           json = [{'id': '5421',
                    'name': 'Jane Green',
                    'emails': [{'home': 'jane@gmail.com'},
                               {'work': 'jane@mywork.com'}
                              ]
                   }
                  ]
           tbl = Table(json)
           print (tbl)
           >>> {'id': '5421', 'name': 'Jane Green', 'emails': [{'home': 'jane@gmail.com'}, {'work': 'jane@mywork.com'}]}
           >>> {'id': '5421', 'name': 'Jane Green', 'emails': [{'home': 'jane@gmail.com'}, {'work': 'jane@mywork.com'}]}

           # Create skinny table of just the nested dicts
           email_skinny = tbl.long_table(['id'], 'emails')

           print (email_skinny)
           >>> {'id': '5421', 'emails_home': 'jane@gmail.com', 'emails_work': None}
           >>> {'id': '5421', 'emails_home': None, 'emails_work': 'jane@mywork.com'}

        `Args:`
            key: lst
                The columns to retain in the long table (e.g. foreign keys)
            column: str
                The column name to make long
            key_rename: dict
                The new name for the foreign key to better identify it. For
                example, you might want to rename ``id`` to ``person_id``.
                Ex. {'KEY_NAME': 'NEW_KEY_NAME'}
            retain_original: boolean
                Retain the original column from the source table.
            prepend:
                Prepend the column name of the unpacked values. Useful for
                avoiding duplicate column names
            prepend_value:
                Value to prepend new columns if ``prepend=True``. If None, will
                set to column name.
        `Returns:`
            Parsons Table
                The new long table
        """

        if type(key) is str:
            key = [key]

        lt = self.cut(*key, column)  # Create a table of key and column
        lt.unpack_list(column, replace=True)  # Unpack the list
        lt.table = petl.melt(lt.table, key)  # Melt into a long table
        lt = lt.cut(*key, "value")  # Get rid of column names created in unpack
        lt.rename_column("value", column)  # Rename 'value' to old column name
        lt.remove_null_rows(column)  # Remove null values

        # If a new key name is specified, rename
        if key_rename:
            for k, v in key_rename.items():
                lt.rename_column(k, v)

        # If there is a nested dict in the column, unpack it
        if lt.num_rows > 0 and isinstance(lt.table[column][0], dict):
            lt.unpack_dict(column, prepend=prepend, prepend_value=prepend_value)

        if not retain_original:
            self.remove_column(column)

        return lt

    def cut(self, *columns):
        r"""
        Return a table of selection of columns

        `Args:`
            *columns: str
                Columns in the parsons table
        `Returns:`
            A new parsons table containing the selected columnns
        """

        from parsons.etl.table import Table

        return Table(petl.cut(self.table, *columns))

    def select_rows(self, *filters):
        r"""
        Select specific rows from a Parsons table based on the passed
        filters.

        Example filters:

        .. code-block:: python

            tbl = Table([['foo', 'bar', 'baz'],
                         ['c', 4, 9.3],
                         ['a', 2, 88.2],
                         ['b', 1, 23.3],])

            # You can structure the filter in multiple wayss

            # Lambda Function
            tbl2 = tbl.select_rows(lambda row: row.foo == 'a' and row.baz > 88.1)
            tbl2
            >>> {'foo': 'a', 'bar': 2, 'baz': 88.1}

            # Expression String
            tbl3 = tbl.select_rows("{foo} == 'a' and {baz} > 88.1")
            tbl3
            >>> {'foo': 'a', 'bar': 2, 'baz': 88.1}

        `Args:`
            *filters: function or str
        `Returns:`
            A new parsons table containing the selected rows
        """

        from parsons.etl.table import Table

        return Table(petl.select(self.table, *filters))

    def remove_null_rows(self, columns, null_value=None):
        """
        Remove rows if the values in a column are ``None``. If multiple columns
        are passed as list, it will remove all rows with null values in any
        of the passed columns.

        `Args:`
            column: str or list
                The column or columns to analyze
            null_value: int or float or str
                The null value
        `Returns:`
            ``None``
        """
        if isinstance(columns, str):
            columns = [columns]

        for col in columns:
            self.table = petl.selectisnot(self.table, col, null_value)

        return self

    def _prepend_dict(self, dict_obj, prepend):
        # Internal method to rename dict keys

        new_dict = {}

        for k, v in dict_obj.items():
            new_dict[prepend + "_" + k] = v

        return new_dict

    def stack(self, *tables, missing=None):
        """
        Stack Parsons tables on top of one another.

        Similar to ``table.concat()``, except no attempt is made to align fields from
        different tables.

        `Args:`
            tables: Parsons Table or list
                A single table, or a list of tables
            missing: bool
                The value to use when padding missing values
        `Returns:`
            ``None``
        """

        if type(tables) not in [list, tuple]:
            tables = [tables]
        petl_tables = [tbl.table for tbl in tables]

        self.table = petl.stack(self.table, *petl_tables, missing=missing)

    def concat(self, *tables, missing=None):
        """
        Concatenates one or more tables onto this one.

        Note that the tables do not need to share exactly the same fields.
        Any missing fields will be padded with None, or whatever is provided via the
        ``missing`` keyword argument.

        `Args:`
            tables: Parsons Table or list
                A single table, or a list of tables
            missing: bool
                The value to use when padding missing values
        `Returns:`
            ``None``
        """

        if type(tables) not in [list, tuple]:
            tables = [tables]
        petl_tables = [tbl.table for tbl in tables]

        self.table = petl.cat(self.table, *petl_tables, missing=missing)

    def chunk(self, rows: int):
        """
        Divides a Parsons table into smaller tables of a specified row count. If the table
        cannot be divided evenly, then the final table will only include the remainder.

        `Args:`
            rows: int
                The number of rows of each new Parsons table
        `Returns:`
            List of Parsons tables
        """

        from parsons.etl import Table

        return [
            Table(petl.rowslice(self.table, i, i + rows)) for i in range(0, self.num_rows, rows)
        ]

    @staticmethod
    def get_normalized_column_name(column_name):
        """
        Returns a column name with whitespace removed, non-alphanumeric characters removed, and
        everything lowercased.

        `Returns:`
            str
                Normalized column name
        """

        column_name = column_name.lower().strip()
        return "".join(c for c in column_name if c.isalnum())

    def match_columns(
        self,
        desired_columns,
        fuzzy_match=True,
        if_extra_columns="remove",
        if_missing_columns="add",
    ):
        """
        Changes the column names and ordering in this Table to match a list of desired column
        names.

        `Args:`
            desired_columns: list
                Ordered list of desired column names
            fuzzy_match: bool
                Whether to normalize column names when matching against the desired column names,
                removing whitespace and non-alphanumeric characters, and lowercasing everything.
                Eg. With this flag set, "FIRST NAME" would match "first_name".
                If the Table has two columns that normalize to the same string (eg. "FIRST NAME"
                and "first_name"), the latter will be considered an extra column.
            if_extra_columns: string
                If the Table has columns that don't match any desired columns, either 'remove'
                them, 'ignore' them, or 'fail' (raising an error).
            if_missing_columns: string
                If the Table is missing some of the desired columns, either 'add' them (with a
                value of None), 'ignore' them, or 'fail' (raising an error).

        `Returns:`
            `Parsons Table` and also updates self
        """

        from parsons.etl import Table  # Just trying to avoid recursive imports.

        normalize_fn = Table.get_normalized_column_name if fuzzy_match else (lambda s: s)

        # Create a mapping of our "normalized" name to the original column name
        current_columns_normalized = {normalize_fn(col): col for col in reversed(self.columns)}

        # Track any columns we need to add to our current table from our desired columns
        columns_to_add = []
        # We are going to do a "cut" later to trim our table and re-order the columns, but
        # we won't have renamed our columns yet, so we need to remember their un-normalized
        # form
        cut_columns = []
        # We are going to also rename our columns AFTER we cut, so we want to remember their
        # normalized names
        final_header = []

        # Loop through our desired columns -- the columns we want to see in our final table
        for desired_column in desired_columns:
            normalized_desired = normalize_fn(desired_column)
            # Try to find our desired column in our Table
            if normalized_desired not in current_columns_normalized:
                # If we can't find our desired column in our current columns, then it's "missing"
                if if_missing_columns == "fail":
                    # If our missing strategy is to fail, raise an exception
                    raise TypeError(f"Table is missing column {desired_column}")
                elif if_missing_columns == "add":
                    # We have to add to our table
                    columns_to_add.append(desired_column)
                    # We will need to remember this column when we cut down to desired columns
                    cut_columns.append(desired_column)
                    # This will be in the final table
                    final_header.append(desired_column)
                elif if_missing_columns != "ignore":
                    # If it's not ignore, add, or fail, then it's not a valid strategy
                    raise TypeError(
                        f"Invalid option {if_missing_columns} for argument `if_missing_columns`"
                    )
            else:
                # We have found this in our current columns, so take it out of our list to search
                current_column = current_columns_normalized.pop(normalized_desired)
                # Add the column to our intermediate table as the old column name
                cut_columns.append(current_column)
                # Add to our final header list as the "desired" name
                final_header.append(desired_column)

        # Look for any "extra" columns from our current table that aren't in our desired columns
        for current_column in current_columns_normalized.values():
            # Figure out what to do with our "extra" columns
            if if_extra_columns == "fail":
                # If our missing strategy is to fail, raise an exception
                raise TypeError(f"Table has extra column {current_column}")
            elif if_extra_columns == "ignore":
                # If we're "ignore"ing our extra columns, we should keep them by adding them to
                # our intermediate and final columns list
                cut_columns.append(current_column)
                final_header.append(current_column)
            elif if_extra_columns != "remove":
                # If it's not ignore, add, or fail, then it's not a valid strategy
                raise TypeError(
                    f"Invalid option {if_extra_columns} for argument `if_extra_columns`"
                )

        # Add any columns we need to add
        for column in columns_to_add:
            self.table = petl.addfield(self.table, column, None)

        # Cut down to just the columns we care about
        self.table = petl.cut(self.table, *cut_columns)

        # Rename any columns
        self.table = petl.setheader(self.table, final_header)

        return self

    def reduce_rows(self, columns, reduce_func, headers, presorted=False, **kwargs):
        """
        Group rows by a column or columns, then reduce the groups to a single row.

        Based on the `rowreduce petl function <https://petl.readthedocs.io/en/stable/transform.html#petl.transform.reductions.rowreduce>`_.

        For example, the output from the query to get a table's definition is
        returned as one component per row. The `reduce_rows` method can be used
        to reduce all those to a single row containg the entire query.

        .. code-block:: python

            >>> ddl = rs.query(sql_to_get_table_ddl)
            >>> ddl.table

            +--------------+--------------+----------------------------------------------------+
            | schemaname   | tablename    | ddl                                                |
            +==============+==============+====================================================+
            | 'db_scratch' | 'state_fips' | '--DROP TABLE db_scratch.state_fips;'              |
            +--------------+--------------+----------------------------------------------------+
            | 'db_scratch' | 'state_fips' | 'CREATE TABLE IF NOT EXISTS db_scratch.state_fips' |
            +--------------+--------------+----------------------------------------------------+
            | 'db_scratch' | 'state_fips' | '('                                                |
            +--------------+--------------+----------------------------------------------------+
            | 'db_scratch' | 'state_fips' | '\\tstate VARCHAR(1024)   ENCODE RAW'               |
            +--------------+--------------+----------------------------------------------------+
            | 'db_scratch' | 'state_fips' | '\\t,stusab VARCHAR(1024)   ENCODE RAW'             |
            +--------------+--------------+----------------------------------------------------+

            >>> reducer_fn = lambda columns, rows: [
            ...     f"{columns[0]}.{columns[1]}",
            ...     '\\n'.join([row[2] for row in rows])]
            >>> ddl.reduce_rows(
            ...     ['schemaname', 'tablename'],
            ...     reducer_fn,
            ...     ['tablename', 'ddl'],
            ...     presorted=True)
            >>> ddl.table

            +-------------------------+-----------------------------------------------------------------------+
            | tablename               | ddl                                                                   |
            +=========================+=======================================================================+
            | 'db_scratch.state_fips' | '--DROP TABLE db_scratch.state_fips;\\nCREATE TABLE IF NOT EXISTS      |
            |                         | db_scratch.state_fips\\n(\\n\\tstate VARCHAR(1024)   ENCODE RAW\\n\\t      |
            |                         | ,db_scratch.state_fips\\n(\\n\\tstate VARCHAR(1024)   ENCODE RAW         |
            |                         | \\n\\t,stusab VARCHAR(1024)   ENCODE RAW\\n\\t,state_name                 |
            |                         | VARCHAR(1024)   ENCODE RAW\\n\\t,statens VARCHAR(1024)   ENCODE         |
            |                         | RAW\\n)\\nDISTSTYLE EVEN\\n;'                                            |
            +-------------------------+-----------------------------------------------------------------------+

        `Args:`
            columns: list
                The column(s) by which to group the rows.
            reduce_func: fun
                The function by which to reduce the rows. Should take the 2
                arguments, the columns list and the rows list and return a list.
                `reducer(columns: list, rows: list) -> list;`
            headers: list
                The list of headers for modified table. The length of `headers`
                should match the length of the list returned by the reduce
                function.
            presorted: bool
                If false, the row will be sorted.
        `Returns:`
            `Parsons Table` and also updates self

        """

        self.table = petl.rowreduce(
            self.table,
            columns,
            reduce_func,
            header=headers,
            presorted=presorted,
            **kwargs,
        )

        return self

    def sort(self, columns=None, reverse=False):
        """
        Sort the rows a table.

        `Args:`
            sort_columns: list or str
                Sort by a single column or a list of column. If ``None`` then
                will sort columns from left to right.
            reverse: boolean
                Sort rows in reverse order.
        `Returns:`
            `Parsons Table` and also updates self
        """

        self.table = petl.sort(self.table, key=columns, reverse=reverse)

        return self

    def set_header(self, new_header):
        """
        Replace the header row of the table.

        `Args:`
            new_header: list
                List of new header column names
        `Returns:`
            `Parsons Table` and also updates self
        """
        self.table = petl.setheader(self.table, new_header)
        return self

    def use_petl(self, petl_method, *args, **kwargs):
        """
        Call a petl function on the current table.

        This convenience method exposes the petl functions to the current
        Table. This is useful in cases where one might need a ``petl`` function
        that has not yet been implemented for ``parsons.Table``.

        .. code-block:: python

            >>> # https://petl.readthedocs.io/en/v1.6.0/transform.html#petl.transform.basics.skipcomments
            >>> tbl = Table([
            ...     ['col1', 'col2'],
            ...     ['# this is a comment row',],
            ...     ['a', 1],
            ...     ['#this is another comment', 'this is also ignored'],
            ...     ['b', 2]
            ... ])
            >>> tbl.use_petl('skipcomments', '#', update_table=True)

            {'col1': 'a', 'col2': 1}
            {'col1': 'b', 'col2': 2}

            >>> tbl.table

            +------+------+
            | col1 | col2 |
            +======+======+
            | 'a'  |    1 |
            +------+------+
            | 'b'  |    2 |
            +------+------+

        `Args:`
            petl_method: str
                The ``petl`` function to call
            update_table: bool
                If ``True``, updates the ``parsons.Table``. Defaults to
                ``False``.
            to_petl: bool
                If ``True``, returns a petl table, otherwise a ``parsons.Table``.
                Defaults to ``False``.
            *args: Any
                The arguements to pass to the petl function.
            **kwargs: Any
                The keyword arguements to pass to the petl function.
        `Returns:`
            `parsons.Table` or `petl` table
        """
        update_table = kwargs.pop("update_table", False)
        to_petl = kwargs.pop("to_petl", False)

        if update_table:
            self.table = getattr(petl, petl_method)(self.table, *args, **kwargs)

        if to_petl:
            return getattr(petl, petl_method)(self.table, *args, **kwargs)

        from parsons.etl.table import Table

        return Table(getattr(petl, petl_method)(self.table, *args, **kwargs))

    def deduplicate(self, keys=None, presorted=False):
        """
        Deduplicates table based on an optional ``keys`` argument,
        which can contain any number of keys or None.

        Method considers all keys specified in the ``keys`` argument
        when deduplicating, not each key individually. For example,
        if ``keys=['a', 'b']``, the method will not remove a record
        unless it's identical to another record in both columns ``a`` and ``b``.

        .. code-block:: python

            >>> tbl = Table([['a', 'b'], [1, 3], [1, 2], [1, 2], [2, 3]])
            >>> tbl.table
            +---+---+
            | a | b |
            +===+===+
            | 1 | 3 |
            +---+---+
            | 1 | 2 |
            +---+---+
            | 1 | 2 |
            +---+---+
            | 2 | 3 |
            +---+---+

            >>> tbl.deduplicate('a')
            >>> # removes all subsequent rows with {'a': 1}
            >>> tbl.table
            +---+---+
            | a | b |
            +===+===+
            | 1 | 3 |
            +---+---+
            | 2 | 3 |
            +---+---+

            >>> tbl = Table([['a', 'b'], [1, 3], [1, 2], [1, 2], [2, 3]]) # reset
            >>> tbl.deduplicate(['a', 'b'])
            >>> # sorted on both ('a', 'b') so (1, 2) was placed before (1, 3)
            >>> # did not remove second instance of {'a': 1} or {'b': 3}
            >>> tbl.table
            +---+---+
            | a | b |
            +===+===+
            | 1 | 2 |
            +---+---+
            | 1 | 3 |
            +---+---+
            | 2 | 3 |
            +---+---+


            >>> tbl = Table([['a', 'b'], [1, 3], [1, 2], [1, 2], [2, 3]]) # reset
            >>> tbl.deduplicate('a').deduplicate('b')
            >>> # can chain method to sort/dedupe on 'a', then sort/dedupe on 'b'
            >>> tbl.table
            +---+---+
            | a | b |
            +===+===+
            | 1 | 3 |
            +---+---+

            >>> tbl = Table([['a', 'b'], [1, 3], [1, 2], [1, 2], [2, 3]]) # reset
            >>> tbl.deduplicate('b').deduplicate('a')
            >>> # Order DOES matter when deduping on one column at a time
            >>> tbl.table
            +---+---+
            | a | b |
            +===+===+
            | 1 | 2 |
            +---+---+

        `Args:`
            keys: str or list[str] or None
                keys to deduplicate (and optionally sort) on.
            presorted: bool
                If false, the row will be sorted.
        `Returns`:
            `Parsons Table` and also updates self

        """

        deduped = petl.transform.dedup.distinct(self.table, key=keys, presorted=presorted)
        self.table = deduped

        return self



import unittest

import requests_mock

import test.test_mobilize.test_mobilize_json as test_json
from parsons import MobilizeAmerica
from test.utils import validate_list


class TestMobilizeAmerica(unittest.TestCase):
    def setUp(self):
        self.ma = MobilizeAmerica()

    def tearDown(self):
        pass

    def test_time_parse(self):
        # Test that Unix conversion works correctly
        self.assertEqual(self.ma._time_parse("<=2018-12-13"), "lte_1544659200")

        # Test that it throws an error when you put in an invalid filter
        self.assertRaises(ValueError, self.ma._time_parse, "=2018-12-01")

    @requests_mock.Mocker()
    def test_get_organizations(self, m):
        m.get(self.ma.uri + "organizations", json=test_json.GET_ORGANIZATIONS_JSON)

        expected = [
            "id",
            "name",
            "slug",
            "is_coordinated",
            "is_independent",
            "is_primary_campaign",
            "state",
            "district",
            "candidate_name",
            "race_type",
            "event_feed_url",
            "created_date",
            "modified_date",
        ]

        # Assert response is expected structure
        self.assertTrue(validate_list(expected, self.ma.get_organizations()))

    @requests_mock.Mocker()
    def test_get_events(self, m):
        m.get(self.ma.uri + "events", json=test_json.GET_EVENTS_JSON)

        expected = [
            "id",
            "description",
            "timezone",
            "title",
            "summary",
            "featured_image_url",
            "event_type",
            "created_date",
            "modified_date",
            "browser_url",
            "high_priority",
            "contact",
            "visibility",
            "sponsor_candidate_name",
            "sponsor_created_date",
            "sponsor_district",
            "sponsor_event_feed_url",
            "sponsor_id",
            "sponsor_is_coordinated",
            "sponsor_is_independent",
            "sponsor_is_primary_campaign",
            "sponsor_modified_date",
            "sponsor_name",
            "sponsor_race_type",
            "sponsor_slug",
            "sponsor_state",
            "address_lines",
            "congressional_district",
            "locality",
            "postal_code",
            "region",
            "state_leg_district",
            "state_senate_district",
            "venue",
            "latitude",
            "longitude",
            "timeslots_0_end_date",
            "timeslots_0_id",
            "timeslots_0_start_date",
        ]

        # Assert response is expected structure
        self.assertTrue(validate_list(expected, self.ma.get_events()))

    @requests_mock.Mocker()
    def test__get_events_organization__can_exclude_timeslots(self, m):
        m.get(requests_mock.ANY, json=test_json.GET_EVENTS_ORGANIZATION_JSON)

        ma = MobilizeAmerica(api_key="test_password")
        data = ma.get_events_organization(1, max_timeslots=0)

        self.assertNotIn("timeslots_0_id", data.columns)

    @requests_mock.Mocker()
    def test__get_events_organization__can_get_all_timeslots(self, m):
        m.get(requests_mock.ANY, json=test_json.GET_EVENTS_ORGANIZATION_JSON)

        ma = MobilizeAmerica(api_key="test_password")
        data = ma.get_events_organization(1)

        self.assertIn("timeslots_0_id", data.columns)
        self.assertIn("timeslots_1_id", data.columns)

    @requests_mock.Mocker()
    def test__get_events_organization__can_limit_timeslots(self, m):
        m.get(requests_mock.ANY, json=test_json.GET_EVENTS_ORGANIZATION_JSON)

        ma = MobilizeAmerica(api_key="test_password")
        data = ma.get_events_organization(1, max_timeslots=1)

        self.assertIn("timeslots_0_id", data.columns)
        self.assertNotIn("timeslots_1_id", data.columns)

    @requests_mock.Mocker()
    def test_get_events_deleted(self, m):
        m.get(self.ma.uri + "events/deleted", json=test_json.GET_EVENTS_DELETED_JSON)

        # Assert response is expected structure
        self.assertTrue(validate_list(["id", "deleted_date"], self.ma.get_events_deleted()))



import datetime
import logging

from parsons import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

logger = logging.getLogger(__name__)

URI = "https://donorbox.org/api/v1"


class Donorbox(object):
    """
    Instantiate Donorbox class.

       `Args:`
            donorbox_account_email: str
                The email associated with your Donorbox account. Can be passed as
                argument or set as ``DONORBOX_ACCOUNT_EMAIL`` environment variable.
            donorbox_api_key: str
                The API key generated by Donorbox for your account. Can be passed as
                argument or set as ``DONORBOX_API_KEY`` environment variable.
    """

    def __init__(self, email=None, api_key=None):
        self.email = check_env.check("DONORBOX_ACCOUNT_EMAIL", email)
        self.api_key = check_env.check("DONORBOX_API_KEY", api_key)
        self.uri = URI
        self.client = APIConnector(self.uri, auth=(self.email, self.api_key))

    def get_campaigns(self, **kwargs):
        """
        Get information on campaigns.

        `Args:`
            id: int or str
                Optional. The ID of the campaign to get. If both id and name are omitted, returns
                all campaigns.
            name: str
                Optional. The name of the campaign to get. If both id and name are omitted, retunrs
                all campaigns.
            order: str
                Optional. Valid values are "asc" and "desc". If not supplied, order is descending
                by default.
            page: int
                Optional. Donorbox supports pagination for larger results. Use the page to track
                your progress.
            per_page: int
                Optional. Results per page when using pagination. Default is 50, maximum is 100.

        `Returns`:
            Parsons Table
        """
        result = self.client.request("campaigns", "GET", params=kwargs)
        data = result.json()
        return Table(data)

    def get_donations(self, **kwargs):
        """
        Get information on donations.

        `Args:`
            email: str
                Optional. Filter's donations by donor's email
            date_from: str
                |  Optional. Filters donations to those started on or after the provided date.
                |  Valid formats: YYYY-mm-dd YYYY/mm/dd YYYYmmdd dd-mm-YYYY
                |  If an incorrectly formatted date is provided, an error is raised.
            date_to: str
                |  Optional. Filters donations to those started before the provided date.
                |  Valid formats: YYYY-mm-dd YYYY/mm/dd YYYYmmdd dd-mm-YYYY
                |  If an incorrectly formatted date is provided, an error is raised.
            campaign_name: str
                Optional. Filters by the campaign title that you have defined in Donorbox.
            campaign_id: int or str
                Optional. Filters by Donorbox campaign id.
            donation_id: int or str
                Optional. Filters by Donorbox donations id.
            first_name: str
                Optional. Filters by donor's first name.
            last_name: str
                Optional. Filters by donor's last name.
            donor_id: int or str
                Optional. Filters by Donorbox donor id.
            amount_min: int or str
                Optional. Gets all donations above the provided minimum.
            amount_max: int or str
                Optional. Gets all donations below the provided maximum.
            order: str
                Optional. Valid values are "asc" and "desc". If not supplied, order is
                descending by default.
            page: int
                Optional. Donorbox supports pagination for larger results. Use the page
                to track your progress.
            per_page: int
                Optional. Results per page when using pagination. Default is 50, maximum
                is 100.

        `Returns`:
            Parsons Table
        """
        # switch variable names
        if "amount_max" in kwargs:
            kwargs["amount[usd][max]"] = kwargs.pop("amount_max")
        if "amount_min" in kwargs:
            kwargs["amount[usd][min]"] = kwargs.pop("amount_min")
        if "donation_id" in kwargs:
            kwargs["id"] = kwargs.pop("donation_id")
        self._check_date_helper(kwargs)
        data = self.client.get_request("donations", params=kwargs)
        return Table(data)

    def get_donors(self, **kwargs):
        """
        Get information on donors.

        `Args:`

            donor_id: str or int
                Optional. Filters by donor ID.
            first_name: str
                Optional. Filters by donor's first name.
            last_name: str
                Optional. Filters by donor's last name.
            donor_name: str
                Optional. Filter by donor's full name
            email: str
                Optional. Filter's donations by donor's email
            order: str
                Optional. Valid values are "asc" and "desc". If not supplied, order is descending
                by default.
            page: int
                Optional. Donorbox supports pagination for larger results. Use the page to track
                your progress.
            per_page: int
                Optional. Results per page when using pagination. Default is 50, maximum is 100.

        `Returns`:
            Parsons Table
        """
        if "donor_id" in kwargs:
            kwargs["id"] = kwargs.pop("donor_id")  # switch to Donorbox's (less specific) name
        data = self.client.get_request("donors", params=kwargs)
        return Table(data)

    def get_plans(self, **kwargs):
        """
        Get information on plans.

        `Args:`

            email: str
                Optional. Filter's plans by donor's email address.
            date_from: str
                |  Optional. Filters plans to those started on or after the provided date.
                |  Valid formats: YYYY-mm-dd YYYY/mm/dd YYYYmmdd dd-mm-YYYY
                |  If an incorrectly formatted date is provided, an error is raised.
            date_to: str
                |  Optional. Filters plans to those started before the provided date.
                |  Valid formats: YYYY-mm-dd YYYY/mm/dd YYYYmmdd dd-mm-YYYY
                |  If an incorrectly formatted date is provided, an error is raised.
            campaign_id: int or str
                Optional. Filters by Donorbox campaign id.
            campaign_name: str
                Optional. Filters by the campaign title that you have defined in Donorbox.
            donor_id: str or int
                Optional. Filters by donor ID.
            first_name: str
                Optional. Filters by donor's first name.
            last_name: str
                Optional. Filters by donor's last name.
            donor_name: str
                Optional. Filter by donor's full name
            order: str
                Optional. Valid values are "asc" and "desc". If not supplied, order is descending
                by default.
            page: int
                Optional. Donorbox supports pagination for larger results. Use the page to track
                your progress.
            per_page: int
                Optional. Results per page when using pagination. Default is 50, maximum is 100.

        `Returns`:
            Parsons Table
        """
        self._check_date_helper(kwargs)
        data = self.client.get_request("plans", params=kwargs)
        return Table(data)

    def _check_date_helper(self, params):
        """Searches through params for a date parameter and if found, calls format helper.

        params: dictionary
            Required. Dictionary of parameters to be passed to endpoint.

        `Returns`: None

        """
        if "date_from" in params and params["date_from"] is not None:
            self._date_format_helper(params["date_from"])
        if "date_to" in params and params["date_to"] is not None:
            self._date_format_helper(params["date_to"])

    def _date_format_helper(self, date_string):
        """Checks date format and warns if invalid (internal)

        Valid formats: YYYY-mm-dd YYYY/mm/dd YYYYmmdd dd-mm-YYYY

        date_string: str
            Required. Date in a string format to be checked against Donorbox's valid options.

        `Returns`: None
        """
        valid_formats = ["%Y-%m-%d", "%d-%m-%Y", "%Y/%m/%d", "%Y%m%d"]
        for str_format in valid_formats:
            try:
                datetime.datetime.strptime(date_string, str_format)
                return
            except ValueError:
                continue
        raise ValueError(
            f"The date you supplied, {date_string}, is not a valid Donorbox format."
            "Try the following formats: YYYY-mm-dd YYYY/mm/dd YYYYmmdd dd-mm-YYYY"
        )



import unittest

import requests_mock

from parsons import Table
from parsons.formstack.formstack import API_URI, Formstack
from test.test_formstack.formstack_json import (
    folder_json,
    form_fields_json,
    form_json,
    form_submissions_json,
    submission_id,
    submission_json,
)

VALID_RESPONSE_STATUS_CODE = 200


class TestFormstack(unittest.TestCase):
    @requests_mock.Mocker()
    def test_get_folders(self, m):
        fs = Formstack(api_token="token")
        m.get(
            API_URI + "/folder",
            status_code=VALID_RESPONSE_STATUS_CODE,
            json=folder_json,
        )
        folders_data = fs.get_folders()
        self.assertIsInstance(folders_data, Table)

    @requests_mock.Mocker()
    def test_get_forms(self, m):
        fs = Formstack(api_token="token")
        m.get(
            API_URI + "/form",
            status_code=VALID_RESPONSE_STATUS_CODE,
            json=form_json,
        )
        forms_data = fs.get_forms()
        self.assertIsInstance(forms_data, Table)

    @requests_mock.Mocker()
    def test_get_submission(self, m):
        fs = Formstack(api_token="token")
        m.get(
            API_URI + f"/submission/{submission_id}",
            status_code=VALID_RESPONSE_STATUS_CODE,
            json=submission_json,
        )
        submission_data = fs.get_submission(submission_id)
        self.assertIsInstance(submission_data, dict)

    @requests_mock.Mocker()
    def test_get_form_submissions(self, m):
        fs = Formstack(api_token="token")
        m.get(
            API_URI + f"/form/{submission_id}/submission",
            status_code=VALID_RESPONSE_STATUS_CODE,
            json=form_submissions_json,
        )
        form_submissions_data = fs.get_form_submissions(submission_id)
        self.assertIsInstance(form_submissions_data, Table)

    @requests_mock.Mocker()
    def test_get_form_fields(self, m):
        fs = Formstack(api_token="token")
        form_id = 123
        m.get(
            API_URI + f"/form/{form_id}/field",
            status_code=VALID_RESPONSE_STATUS_CODE,
            json=form_fields_json,
        )
        form_fields_data = fs.get_form_fields(form_id)
        self.assertIsInstance(form_fields_data, Table)



# Provide shortcuts to importing Parsons submodules and set up logging
import importlib
import logging
import os

from parsons.etl.table import Table

# Define the default logging config for Parsons and its submodules. For now the
# logger gets a StreamHandler by default. At some point a NullHandler may be more
# appropriate, so the end user must decide on logging behavior.

logger = logging.getLogger(__name__)
_handler = logging.StreamHandler()
_formatter = logging.Formatter("%(module)s %(levelname)s %(message)s")
_handler.setFormatter(_formatter)
logger.addHandler(_handler)

if os.environ.get("TESTING"):
    # Log less stuff in automated tests
    logger.setLevel("WARNING")
elif os.environ.get("DEBUG"):
    logger.setLevel("DEBUG")
else:
    logger.setLevel("INFO")

# Table is referenced by many connectors, so we add it immediately to limit the damage
# of circular dependencies
__all__ = ["Table"]
for module_path, connector_name in (
    ("parsons.actblue.actblue", "ActBlue"),
    ("parsons.action_kit.action_kit", "ActionKit"),
    ("parsons.action_builder.action_builder", "ActionBuilder"),
    ("parsons.action_network.action_network", "ActionNetwork"),
    ("parsons.airmeet.airmeet", "Airmeet"),
    ("parsons.airtable.airtable", "Airtable"),
    ("parsons.alchemer.alchemer", "Alchemer"),
    ("parsons.alchemer.alchemer", "SurveyGizmo"),
    ("parsons.auth0.auth0", "Auth0"),
    ("parsons.aws.s3", "S3"),
    ("parsons.azure.azure_blob_storage", "AzureBlobStorage"),
    ("parsons.bill_com.bill_com", "BillCom"),
    ("parsons.bloomerang.bloomerang", "Bloomerang"),
    ("parsons.box.box", "Box"),
    ("parsons.braintree.braintree", "Braintree"),
    ("parsons.capitol_canary.capitol_canary", "CapitolCanary"),
    ("parsons.catalist.catalist", "CatalistMatch"),
    ("parsons.census.census", "Census"),
    ("parsons.civis.civisclient", "CivisClient"),
    ("parsons.community.community", "Community"),
    ("parsons.controlshift.controlshift", "Controlshift"),
    ("parsons.copper.copper", "Copper"),
    ("parsons.crowdtangle.crowdtangle", "CrowdTangle"),
    ("parsons.databases.database_connector", "DatabaseConnector"),
    ("parsons.databases.discover_database", "discover_database"),
    ("parsons.databases.db_sync", "DBSync"),
    ("parsons.databases.mysql.mysql", "MySQL"),
    ("parsons.databases.postgres.postgres", "Postgres"),
    ("parsons.databases.redshift.redshift", "Redshift"),
    ("parsons.databases.sqlite.sqlite", "Sqlite"),
    ("parsons.donorbox.donorbox", "Donorbox"),
    ("parsons.facebook_ads.facebook_ads", "FacebookAds"),
    ("parsons.formstack.formstack", "Formstack"),
    ("parsons.freshdesk.freshdesk", "Freshdesk"),
    ("parsons.geocode.census_geocoder", "CensusGeocoder"),
    ("parsons.github.github", "GitHub"),
    ("parsons.google.google_admin", "GoogleAdmin"),
    ("parsons.google.google_bigquery", "GoogleBigQuery"),
    ("parsons.google.google_civic", "GoogleCivic"),
    ("parsons.google.google_cloud_storage", "GoogleCloudStorage"),
    ("parsons.google.google_sheets", "GoogleSheets"),
    ("parsons.hustle.hustle", "Hustle"),
    ("parsons.mailchimp.mailchimp", "Mailchimp"),
    ("parsons.mobilecommons.mobilecommons", "MobileCommons"),
    ("parsons.mobilize_america.ma", "MobilizeAmerica"),
    ("parsons.nation_builder.nation_builder", "NationBuilder"),
    ("parsons.newmode.newmode", "Newmode"),
    ("parsons.ngpvan.van", "VAN"),
    ("parsons.notifications.gmail", "Gmail"),
    ("parsons.notifications.slack", "Slack"),
    ("parsons.notifications.smtp", "SMTP"),
    ("parsons.pdi.pdi", "PDI"),
    ("parsons.phone2action.p2a", "Phone2Action"),
    ("parsons.quickbase.quickbase", "Quickbase"),
    ("parsons.quickbooks.quickbookstime", "QuickBooksTime"),
    ("parsons.redash.redash", "Redash"),
    ("parsons.rockthevote.rtv", "RockTheVote"),
    ("parsons.salesforce.salesforce", "Salesforce"),
    ("parsons.scytl.scytl", "Scytl"),
    ("parsons.sftp.sftp", "SFTP"),
    ("parsons.shopify.shopify", "Shopify"),
    ("parsons.sisense.sisense", "Sisense"),
    ("parsons.targetsmart.targetsmart_api", "TargetSmartAPI"),
    ("parsons.targetsmart.targetsmart_automation", "TargetSmartAutomation"),
    ("parsons.turbovote.turbovote", "TurboVote"),
    ("parsons.twilio.twilio", "Twilio"),
    ("parsons.zoom.zoom", "Zoom"),
    ("parsons.empower.empower", "Empower"),
):
    try:
        globals()[connector_name] = getattr(importlib.import_module(module_path), connector_name)
        __all__.append(connector_name)
    except ImportError:
        logger.debug(f"Could not import {module_path}.{connector_name}; skipping")



import json
import logging
import os

from simple_salesforce import Salesforce as _Salesforce

from parsons.utilities import check_env

logger = logging.getLogger(__name__)


class Salesforce:
    """
    Instantiate the Salesforce class

    Supports the password and `client_credentials <https://help.salesforce.com/s/articleView?id=xcloud.connected_app_client_credentials_setup.htm&type=5>`_ authentication methods.

    `Args:`
        username: str
            The Salesforce username (usually an email address). Not required if
            ``SALESFORCE_USERNAME`` env variable is passed. Used in the 'password' auth method.
        password: str
            The Salesforce password. Not required if ``SALESFORCE_PASSWORD`` env variable is
            passed. Used in the 'password' auth method.
        security_token: str
            The Salesforce security token that can be acquired or reset in
            Settings > My Personal Information > Reset My Security Token.
            Not required if ``SALESFORCE_SECURITY_TOKEN`` env variable is passed. Used in the 'password' auth method.
        test_environment: bool
            If ``True`` the client will connect to a Salesforce sandbox instance. Not required if
            ``SALESFORCE_DOMAIN`` env variable is passed.
        consumer_key: str
            consumer key for a connected app. Used in the 'client_credentials' auth method.
        consumer_secret: str
            consumer secret for a connected app. Used in the 'client_credentials' auth method.
        domain: str
            url for the salesforce instance. Used in the 'client_credentials' auth method
        authentication_method: str
            the method to use for authentication. defaults to "password". Not required if ``SALESFORCE_AUTHENTICATION_METHOD`` env variable is passed.

    `Returns:`
        Salesforce class
    """

    def __init__(
        self,
        username=None,
        password=None,
        security_token=None,
        test_environment=False,
        consumer_key=None,
        consumer_secret=None,
        domain=None,
        authentication_method=None,
    ):
        if authentication_method:
            self.authentication_method = authentication_method
        elif env_authentication_method := os.environ.get("SALESFORCE_AUTHENTICATION_METHOD"):
            self.authentication_method = env_authentication_method
        else:
            self.authentication_method = "password"

        if self.authentication_method == "password":
            self.username = check_env.check("SALESFORCE_USERNAME", username)
            self.password = check_env.check("SALESFORCE_PASSWORD", password)
            self.security_token = check_env.check("SALESFORCE_SECURITY_TOKEN", security_token)
            if test_environment:
                self.domain = check_env.check("SALESFORCE_DOMAIN", "test")
            else:
                self.domain = None

        elif self.authentication_method == "client_credentials":
            self.consumer_key = check_env.check("SALESFORCE_CONSUMER_KEY", consumer_key)
            self.consumer_secret = check_env.check("SALESFORCE_CONSUMER_SECRET", consumer_key)
            self.domain = check_env.check("SALESFORCE_DOMAIN", domain)

        else:
            raise NotImplementedError(
                f"{self.authentication_method} is not a supported method. Parsons currently supports 'password' and 'client_credentials'"
            )

        self._client = None

    def describe_object(self, object):
        """
        `Args:`
            object: str
                The API name of the type of record to describe. Note that custom object names end
                in `__c`
        `Returns:`
            Ordered Dict of all the object's meta data in Salesforce
        """

        return getattr(self.client, object).describe()

    def describe_fields(self, object):
        """
        `Args:`
            object: str
                The API name of the type of record on whose fields you want data. Note that custom
                object names end in `__c`
        `Returns:`
            Dict of all the object's field meta data in Salesforce
        """

        return json.loads(json.dumps(getattr(self.client, object).describe()["fields"]))

    def query(self, soql):
        """
        `Args:`
            soql: str
                The desired query in Salesforce SOQL language (SQL with additional limitations).
                For reference, see the `Salesforce SOQL documentation <https://developer.salesforce.com/docs/atlas.en-us.soql_sosl.meta/soql_sosl/sforce_api_calls_soql.htm>`_.
        `Returns:`
            list of dicts with Salesforce data
        """

        q = self.client.query_all(soql)
        q = json.loads(json.dumps(q))
        logger.info(f"Found {q['totalSize']} results")
        return q

    def insert_record(self, object, data_table):
        """
        Insert new records of the desired object into Salesforce

        `Args:`
            object: str
                The API name of the type of record to insert. Note that custom object names end
                in `__c`
            data_table: obj
                A Parsons Table with data for inserting records. Column names must match object
                field API names, though case and order need not match. Note that custom field
                names end in `__c`.
        `Returns:`
            list of dicts that have the following data:
            * success: boolean
            * created: boolean (if new record is created)
            * id: str (id of record created, if successful)
            * errors: list of dicts (with error details)
        """

        r = getattr(self.client.bulk, object).insert(data_table.to_dicts())
        s = [x for x in r if x.get("success") is True]
        logger.info(
            f"Successfully inserted {len(s)} out of {data_table.num_rows} records to {object}"
        )
        return r

    def update_record(self, object, data_table):
        """
        Update existing records of the desired object in Salesforce

        `Args:`
            object: str
                The API name of the type of record to update. Note that custom object names end
                in `__c`
            data_table: obj
                A Parsons Table with data for updating records. Must contain one column named
                `id`. Column names must match object field API names, though case and order need
                not match. Note that custom field names end in `__c`.
            `Returns:`
                list of dicts that have the following data:
                * success: boolean
                * created: boolean (if new record is created)
                * id: str (id of record altered, if successful)
                * errors: list of dicts (with error details)
        """

        r = getattr(self.client.bulk, object).update(data_table.to_dicts())
        s = [x for x in r if x.get("success") is True]
        logger.info(
            f"Successfully updated {len(s)} out of {data_table.num_rows} records in {object}"
        )
        return r

    def upsert_record(self, object, data_table, id_col):
        """
        Insert new records and update existing ones of the desired object in Salesforce

        `Args:`
            object: str
                The API name of the type of record to upsert. Note that custom object names end
                in `__c`
            data_table: obj
                A Parsons Table with data for upserting records. Column names must match object
                field API names, though case and order need not match. Note that custom field
                names end in `__c`.
            id_col: str
                The column name in `data_table` that stores the record ID. Required even if all
                records are new/inserted.
            `Returns:`
                list of dicts that have the following data:
                * success: boolean
                * created: boolean (if new record is created)
                * id: str (id of record created or altered, if successful)
                * errors: list of dicts (with error details)
        """

        r = getattr(self.client.bulk, object).upsert(data_table.to_dicts(), id_col)
        s = [x for x in r if x.get("success") is True]
        logger.info(
            f"Successfully upserted {len(s)} out of {data_table.num_rows} records to {object}"
        )
        return r

    def delete_record(self, object, id_table, hard_delete=False):
        """
        Delete existing records of the desired object in Salesforce

        `Args:`
            object: str
                The API name of the type of record to delete. Note that custom object names end
                in `__c`
            id_table: obj
                A Parsons Table of record IDs to delete. Note that 'Id' is the default Salesforce
                record ID field name.
            hard_delete: boolean
                If true, will permanently delete record instead of moving it to trash
            `Returns:`
                list of dicts that have the following data:
                * success: boolean
                * created: boolean (if new record is created)
                * id: str (id of record deleted, if successful)
                * errors: list of dicts (with error details)
        """

        if hard_delete:
            r = getattr(self.client.bulk, object).hard_delete(id_table.to_dicts())
        else:
            r = getattr(self.client.bulk, object).delete(id_table.to_dicts())

        s = [x for x in r if x.get("success") is True]
        logger.info(
            f"Successfully deleted {len(s)} out of {id_table.num_rows} records from {object}"
        )
        return r

    @property
    def client(self):
        """
        Get the Salesforce client to use for making all calls. For more information, check the
        `Simple Salesforce Documentation <https://simple-salesforce.readthedocs.io/en/latest/>`_

        `Returns:`
            `simple-salesforce Salesforce object`
        """
        if not self._client:
            # Create a Salesforce client to use to make bulk calls
            if self.authentication_method == "password":
                self._client = _Salesforce(
                    username=self.username,
                    password=self.password,
                    security_token=self.security_token,
                    domain=self.domain,
                )
            elif self.authentication_method == "client_credentials":
                self._client = _Salesforce(
                    consumer_key=self.consumer_key,
                    consumer_secret=self.consumer_secret,
                    domain=self.domain,
                )
            else:
                raise Exception("Should not be possible to reach this code")

        return self._client



import os
import unittest

from parsons import MySQL, Table
from parsons.databases.mysql.create_table import MySQLCreateTable
from test.utils import assert_matching_tables


# These tests interact directly with the MySQL database. To run, set env variable "LIVE_TEST=True"
@unittest.skipIf(not os.environ.get("LIVE_TEST"), "Skipping because not running live test")
class TestMySQLLive(unittest.TestCase):
    def setUp(self):
        self.mysql = MySQL()

    def tearDown(self):
        # Drop the view, the table and the schema
        sql = "DROP TABLE IF EXISTS test;"
        self.mysql.query(sql)

    def test_query(self):
        # Check that query sending back expected result
        r = self.mysql.query("SELECT 1")
        self.assertEqual(r.first, 1)

    def test_query_no_response(self):
        # Check that a query that has no response and doesn't fail
        sql = "CREATE TABLE test (name VARCHAR(255), user_name VARCHAR(255))"
        r = self.mysql.query(sql)
        self.assertEqual(r, None)

    def test_insert_data(self):
        sql = "CREATE TABLE test (name VARCHAR(255), user_name VARCHAR(255));"
        self.mysql.query(sql)

        sql = "INSERT INTO test (name, user_name) VALUES ('me', 'myuser');"
        self.mysql.query(sql)

        r = self.mysql.query("select * from test")

        assert_matching_tables(Table([{"name": "me", "user_name": "myuser"}]), r)


# These tests interact directly with the MySQL database. To run, set env variable "LIVE_TEST=True"
@unittest.skipIf(not os.environ.get("LIVE_TEST"), "Skipping because not running live test")
class TestMySQL(unittest.TestCase):
    def setUp(self):
        self.mysql = MySQL()

        # Create tables
        self.mysql.query(
            "CREATE TABLE IF NOT EXISTS test (name VARCHAR(255), user_name VARCHAR(255), id INT)"
        )
        self.mysql.query(
            """
                         INSERT INTO test (name, user_name, id)
                         VALUES ('me', 'myuser', '1'),
                                ('you', 'hey', '2'),
                                ('you', 'hey', '3')
                         """
        )

        self.tbl = MySQLCreateTable(self.mysql, "test")

    def tearDown(self):
        self.mysql.query("DROP TABLE IF EXISTS test;")

    def test_num_rows(self):
        self.assertEqual(self.tbl.num_rows, 3)

    def test_max_primary_key(self):
        self.assertEqual(self.tbl.max_primary_key("id"), 3)

    def test_distinct_primary_key(self):
        self.assertTrue(self.tbl.distinct_primary_key("id"))
        self.assertFalse(self.tbl.distinct_primary_key("user_name"))

    def test_columns(self):
        self.assertEqual(self.tbl.columns, ["name", "user_name", "id"])

    def test_exists(self):
        self.assertTrue(self.tbl.exists)

        tbl_bad = MySQLCreateTable(self.mysql, "bad_test")
        self.assertFalse(tbl_bad.exists)

    def test_drop(self):
        self.tbl.drop()
        self.assertFalse(self.tbl.exists)

    def test_truncate(self):
        self.tbl.truncate()
        self.assertEqual(self.tbl.num_rows, 0)

    def test_get_rows(self):
        data = [
            ["name", "user_name", "id"],
            ["me", "myuser", "1"],
            ["you", "hey", "2"],
            ["you", "hey", "3"],
        ]
        tbl = Table(data)

        assert_matching_tables(self.tbl.get_rows(), tbl)

    def test_get_new_rows(self):
        data = [["name", "user_name", "id"], ["you", "hey", "2"], ["you", "hey", "3"]]
        tbl = Table(data)

        # Basic
        assert_matching_tables(self.tbl.get_new_rows("id", 1), tbl)

        # Chunking
        assert_matching_tables(self.tbl.get_new_rows("id", 1, chunk_size=1), tbl)

    def test_get_new_rows_count(self):
        self.assertEqual(self.tbl.get_new_rows_count("id", 1), 2)


# TODO: figure out why there are 2 of these
class TestMySQL(unittest.TestCase):  # noqa: F811
    def setUp(self):
        self.mysql = MySQL(username="test", password="test", host="test", db="test", port=123)

        self.tbl = Table(
            [
                ["ID", "Name", "Score"],
                [1, "Jim", 1.9],
                [2, "John", -0.5],
                [3, "Sarah", 0.0004],
            ]
        )

    def test_data_type(self):
        # Test bool
        self.assertEqual(self.mysql.data_type(False, ""), "bool")
        self.assertEqual(self.mysql.data_type(True, ""), "bool")

        # Test smallint
        self.assertEqual(self.mysql.data_type(1, ""), "smallint")
        self.assertEqual(self.mysql.data_type(2, ""), "smallint")
        # Test int
        self.assertEqual(self.mysql.data_type(32769, ""), "mediumint")
        # Test bigint
        self.assertEqual(self.mysql.data_type(2147483648, ""), "bigint")
        # Test varchar that looks like an int
        self.assertEqual(self.mysql.data_type("00001", ""), "varchar")
        # Test a float as a decimal
        self.assertEqual(self.mysql.data_type(5.001, ""), "float")
        # Test varchar
        self.assertEqual(self.mysql.data_type("word", ""), "varchar")
        # Test int with underscore as string
        self.assertEqual(self.mysql.data_type("1_2", ""), "varchar")
        # Test int with underscore
        self.assertEqual(self.mysql.data_type(1_2, ""), "smallint")
        # Test int with leading zero
        self.assertEqual(self.mysql.data_type("01", ""), "varchar")

    def test_evaluate_table(self):
        table_map = [
            {"name": "ID", "type": "smallint", "width": 0},
            {"name": "Name", "type": "varchar", "width": 8},
            {"name": "Score", "type": "float", "width": 0},
        ]
        self.assertEqual(self.mysql.evaluate_table(self.tbl), table_map)

    def test_create_statement(self):
        stmt = "CREATE TABLE test_table ( \n id smallint \n,name varchar(10) \n,score float \n);"
        self.assertEqual(self.mysql.create_statement(self.tbl, "test_table"), stmt)



import logging
import re

from parsons.etl import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

logger = logging.getLogger(__name__)

PAGE_SIZE = 100


class Freshdesk:
    """
    Instantiate Freshdesk class

    `Args:`
        domain: str
            The subdomain of the Freshdesk account. Not required if ``FRESHDESK_DOMAIN``
            env variable set.
        api_key: str
            The Freshdesk provided application key. Not required if ``FRESHDESK_API_KEY``
            env variable set.
    `Returns:`
        Freshdesk class
    """

    def __init__(self, domain, api_key):
        self.api_key = check_env.check("FRESHDESK_API_KEY", api_key)
        self.domain = check_env.check("FRESHDESK_DOMAIN", domain)
        self.uri = f"https://{self.domain}.freshdesk.com/api/v2/"
        self.client = APIConnector(self.uri, auth=(self.api_key, "x"))

    def _get_request(self, endpoint, params=None):
        base_params = {"per_page": PAGE_SIZE}

        if params:
            base_params.update(params)

        r = self.client.request(endpoint, "GET", params=base_params)
        self.client.validate_response(r)
        data = r.json()

        # Paginate
        while "link" in r.headers.keys():
            logger.info(f"Retrieving another page of {PAGE_SIZE} records.")
            url = re.search("<(.*)>", r.headers["link"]).group(1)
            r = self.client.request(url, "GET", params=params)
            self.client.validate_response(r)
            data.extend(r.json())

        return data

    def _post_request(self, endpoint, data):
        """
        Send a POST request to the specified Freshdesk endpoint.
        `Args:`
            endpoint: str
                The endpoint of the Freshdesk API to which the request is being sent.
            data: dict
                The data to be sent in the request body.
        `Returns:`
            dict
                The JSON response from the API.
        """
        url = self.uri + endpoint
        r = self.client.request(url, "POST", json=data)
        self.client.validate_response(r)
        return r.json()

    @staticmethod
    def _transform_table(tbl, expand_custom_fields=None):
        if tbl.num_rows > 0:
            tbl.move_column("id", 0)
            tbl.sort()
            if expand_custom_fields:
                tbl.unpack_dict("custom_fields", prepend=False)

        return tbl

    def get_tickets(
        self,
        ticket_type=None,
        requester_id=None,
        requester_email=None,
        company_id=None,
        updated_since="2016-01-01",
        expand_custom_fields=False,
    ):
        """
        List tickets.

        See the `API Docs <https://developers.freshdesk.com/api/#list_all_tickets>`_
        for more information.

        .. warning::
            Deleted and Spam tickets are not included. However they can be pulled separately
            by utilizing the ``ticket_type`` parameter.

        .. warning::
            Freshdesk will return a maximum of 9,000 tickets. By default, only tickets created in
            the past 30 days are returned. To access additional tickets, utilize the
            ``updated_since`` parameter.

        `Args:`
            ticket_type: str
                Filter by type of ticket to filter by. Valid fields include ``new_and_my_open``,
                ``watching``, ``spam`` and ``deleted``.
            requester_id: int
                Filter by requester id.
            requester_email: str
                Filter by requester email.
            company_id: int
                Filter by company_id.
            updated_since: str
                Earliest date to include in results.
            expand_custom_fields: boolean
                Expand nested custom fields to their own columns.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        params = {
            "filter": ticket_type,
            "requester_id": requester_id,
            "requester_email": requester_email,
            "company_id": company_id,
            "updated_since": updated_since,
        }

        tbl = Table(self._get_request("tickets", params=params))
        logger.info(f"Found {tbl.num_rows} tickets.")
        return self._transform_table(tbl, expand_custom_fields)

    def get_contacts(
        self,
        email=None,
        mobile=None,
        phone=None,
        company_id=None,
        state=None,
        updated_since=None,
        expand_custom_fields=None,
    ):
        """
        Get contacts.

        See the `API Docs <https://developers.freshdesk.com/api/#list_all_contacts>`_
        for more information.

        `Args:`
            email: str
                Filter by email address.
            mobile: str
                Filter by mobile phone number.
            phone: str
                Filter by phone number.
            company_id: int
                Filter by company ID.
            state: str
                Filter by state.
            updated_since: str
                Earliest date to include in results.
            expand_custom_fields: boolean
                Expand nested custom fields to their own columns.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        params = {
            "email": email,
            "mobile": mobile,
            "phone": phone,
            "company_id": company_id,
            "state": state,
            "_updated_since": updated_since,
        }

        tbl = Table(self._get_request("contacts", params=params))
        logger.info(f"Found {tbl.num_rows} contacts.")
        return self._transform_table(tbl, expand_custom_fields)

    def get_companies(self, expand_custom_fields=False):
        """
        List companies.

        See the `API Docs <https://developers.freshdesk.com/api/#list_all_companies>`_
        for more information.

        `Args:`
            expand_custom_fields: boolean
                Expand nested custom fields to their own columns.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self._get_request("companies"))
        logger.info(f"Found {tbl.num_rows} companies.")
        return self._transform_table(tbl, expand_custom_fields)

    def get_agents(self, email=None, mobile=None, phone=None, state=None):
        """
        List agents.

        See the `API Docs <https://developers.freshdesk.com/api/#list_all_agents>`_
        for more information.

        `Args:`
            email: str
                Filter by email address.
            mobile: str
                Filter by mobile phone number
            phone: str
                Filter by phone number
            state: str
                Filter by state
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        params = {"email": email, "mobile": mobile, "phone": phone, "state": state}
        tbl = Table(self._get_request("agents", params=params))
        logger.info(f"Found {tbl.num_rows} agents.")
        tbl = self._transform_table(tbl)
        tbl = tbl.unpack_dict("contact", prepend=False)
        tbl.remove_column("signature")  # Removing since raw HTML might cause issues.

        return tbl

    def create_ticket(
        self, subject, description, email, priority, status, cc_emails=None, custom_fields=None
    ):
        """
        Create a ticket in Freshdesk.
        `Args:`
            subject: str
                The subject of the ticket.
            description: str
                The description of the ticket.
            email: str
                The email address of the requester.
            priority: int
                The priority of the ticket.
            status: int
                The status of the ticket.
            cc_emails: list (optional)
                List of email addresses to CC.
            custom_fields: dict (optional)
                Custom fields data.
        `Returns:`
            dict
                JSON response from the API.
        """
        endpoint = "tickets"
        data = {
            "subject": subject,
            "description": description,
            "email": email,
            "priority": priority,
            "status": status,
            "cc_emails": cc_emails if cc_emails else [],
            "custom_fields": custom_fields if custom_fields else {},
        }
        return self._post_request(endpoint, data)



import base64
import email
import json
import os
import shutil
import tempfile
import unittest

import requests_mock

from parsons import Gmail

_dir = os.path.dirname(__file__)


class TestGmail(unittest.TestCase):
    @requests_mock.Mocker()
    def setUp(self, m):
        self.tmp_folder = tempfile.mkdtemp()
        self.credentials_file = f"{self.tmp_folder}/credentials.json"
        self.token_file = f"{self.tmp_folder}/token.json"

        with open(self.credentials_file, "w") as f:
            f.write(
                json.dumps(
                    {
                        "installed": {
                            "client_id": "someclientid.apps.googleusercontent.com",
                            "project_id": "some-project-id-12345",
                            "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                            "token_uri": "https://www.googleapis.com/oauth2/v3/token",
                            "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
                            "client_secret": "someclientsecret",
                            "redirect_uris": [
                                "urn:ietf:wg:oauth:2.0:oob",
                                "http://localhost",
                            ],
                        }
                    }
                )
            )

        with open(self.token_file, "w") as f:
            f.write(
                json.dumps(
                    {
                        "access_token": "someaccesstoken",
                        "client_id": "some-client-id.apps.googleusercontent.com",
                        "client_secret": "someclientsecret",
                        "refresh_token": "1/refreshrate",
                        "token_expiry": "2030-02-20T23:28:09Z",
                        "token_uri": "https://www.googleapis.com/oauth2/v3/token",
                        "user_agent": None,
                        "revoke_uri": "https://oauth2.googleapis.com/revoke",
                        "id_token": None,
                        "id_token_jwt": None,
                        "token_response": {
                            "access_token": "someaccesstoken",
                            "expires_in": 3600000,
                            "scope": "https://www.googleapis.com/auth/gmail.send",
                            "token_type": "Bearer",
                        },
                        "scopes": ["https://www.googleapis.com/auth/gmail.send"],
                        "token_info_uri": "https://oauth2.googleapis.com/tokeninfo",
                        "invalid": False,
                        "_class": "OAuth2Credentials",
                        "_module": "oauth2client.client",
                    }
                )
            )

        self.gmail = Gmail(self.credentials_file, self.token_file)

    def tearDown(self):
        # Delete tmp folder and files
        shutil.rmtree(self.tmp_folder)

    def test_create_message_simple(self):
        sender = "Sender <sender@email.com>"
        to = "Recepient <recepient@email.com>"
        subject = "This is a test email"
        message_text = "The is the message text of the email"

        msg = self.gmail._create_message_simple(sender, to, subject, message_text)
        raw = self.gmail._encode_raw_message(msg)

        decoded = email.message_from_bytes(base64.urlsafe_b64decode(bytes(raw["raw"], "utf-8")))

        expected_items = [
            ("Content-Type", 'text/plain; charset="us-ascii"'),
            ("MIME-Version", "1.0"),
            ("Content-Transfer-Encoding", "7bit"),
            ("to", to),
            ("from", sender),
            ("subject", subject),
        ]

        # Check the metadata
        self.assertListEqual(decoded.items(), expected_items)

        # Check the message
        self.assertEqual(decoded.get_payload(), message_text)

        # Check the number of parts
        expected_parts = 1
        self.assertEqual(sum([1 for i in decoded.walk()]), expected_parts)

    def test_create_message_html(self):
        sender = "Sender <sender@email.com>"
        to = "Recepient <recepient@email.com>"
        subject = "This is a test html email"
        message_text = "The is the message text of the email"
        message_html = "<p>This is the html message part of the email</p>"

        msg = self.gmail._create_message_html(sender, to, subject, message_text, message_html)
        raw = self.gmail._encode_raw_message(msg)

        decoded = email.message_from_bytes(base64.urlsafe_b64decode(bytes(raw["raw"], "utf-8")))

        expected_items = [
            ("Content-Type", "multipart/alternative;\n boundary="),
            ("MIME-Version", "1.0"),
            ("subject", subject),
            ("from", sender),
            ("to", to),
        ]

        # The boundary id changes everytime. Replace it with the beginnig to
        # avoid failures
        updated_items = []
        for i in decoded.items():
            if "Content-Type" in i[0] and "multipart/alternative;\n boundary=" in i[1]:
                updated_items.append(("Content-Type", "multipart/alternative;\n boundary="))
            else:
                updated_items.append((i[0], i[1]))

        # Check the metadata
        self.assertListEqual(updated_items, expected_items)

        # Check the message
        # The first part is just a container for the text and html parts
        parts = decoded.get_payload()

        self.assertEqual(parts[0].get_payload(), message_text)
        self.assertEqual(parts[1].get_payload(), message_html)

        # Check the number of parts
        expected_parts = 3
        self.assertEqual(sum([1 for i in decoded.walk()]), expected_parts)

    def test_create_message_html_no_text(self):
        sender = "Sender <sender@email.com>"
        to = "Recepient <recepient@email.com>"
        subject = "This is a test html email"
        message_html = "<p>This is the html message part of the email</p>"

        msg = self.gmail._create_message_html(sender, to, subject, "", message_html)
        raw = self.gmail._encode_raw_message(msg)

        decoded = email.message_from_bytes(base64.urlsafe_b64decode(bytes(raw["raw"], "utf-8")))

        expected_items = [
            ("Content-Type", "multipart/alternative;\n boundary="),
            ("MIME-Version", "1.0"),
            ("subject", subject),
            ("from", sender),
            ("to", to),
        ]

        # The boundary id changes everytime. Replace it with the beginnig to
        # avoid failures
        updated_items = []
        for i in decoded.items():
            if "Content-Type" in i[0] and "multipart/alternative;\n boundary=" in i[1]:
                updated_items.append(("Content-Type", "multipart/alternative;\n boundary="))
            else:
                updated_items.append((i[0], i[1]))

        # Check the metadata
        self.assertListEqual(updated_items, expected_items)

        # Check the message
        # The first part is just a container for the text and html parts
        parts = decoded.get_payload()

        self.assertEqual(parts[0].get_payload(), message_html)

        # Check the number of parts
        expected_parts = 2
        self.assertEqual(sum([1 for i in decoded.walk()]), expected_parts)

    def test_create_message_attachments(self):
        sender = "Sender <sender@email.com>"
        to = "Recepient <recepient@email.com>"
        subject = "This is a test email with attachements"
        message_text = "The is the message text of the email with attachments"
        message_html = "<p>This is the html message part of the email with attachments</p>"
        attachments = [f"{_dir}/assets/loremipsum.txt"]

        msg = self.gmail._create_message_attachments(
            sender, to, subject, message_text, attachments, message_html=message_html
        )
        raw = self.gmail._encode_raw_message(msg)

        decoded = email.message_from_bytes(base64.urlsafe_b64decode(bytes(raw["raw"], "utf-8")))

        expected_items = [
            ("Content-Type", "multipart/alternative;\n boundary="),
            ("MIME-Version", "1.0"),
            ("to", to),
            ("from", sender),
            ("subject", subject),
        ]

        # The boundary id changes everytime. Replace it with the beginnig to
        # avoid failures
        updated_items = []
        for i in decoded.items():
            if "Content-Type" in i[0] and "multipart/alternative;\n boundary=" in i[1]:
                updated_items.append(("Content-Type", "multipart/alternative;\n boundary="))
            else:
                updated_items.append((i[0], i[1]))

        # Check the metadata
        self.assertListEqual(updated_items, expected_items)

        # Check the message
        # The first part is just a container for the text and html parts
        parts = decoded.get_payload()

        self.assertEqual(parts[0].get_payload(), message_text)
        self.assertEqual(parts[1].get_payload(), message_html)

        if os.linesep == "\r\n":
            file = f"{_dir}/assets/loremipsum_b64_win_txt.txt"
        else:
            file = f"{_dir}/assets/loremipsum_b64_txt.txt"

        with open(file, "r") as f:
            b64_txt = f.read()
        self.assertEqual(parts[2].get_payload(), b64_txt)

        self.assertEqual(parts[2].get_content_type(), "text/plain")

        # Check the number of parts
        expected_parts = 4
        self.assertEqual(sum([1 for i in decoded.walk()]), expected_parts)

    def test_create_message_attachments_jpeg(self):
        sender = "Sender <sender@email.com>"
        to = "Recepient <recepient@email.com>"
        subject = "This is a test email with attachements"
        message_text = "The is the message text of the email with attachments"
        message_html = "<p>This is the html message part of the email with attachments</p>"
        attachments = [f"{_dir}/assets/loremipsum.jpeg"]

        msg = self.gmail._create_message_attachments(
            sender, to, subject, message_text, attachments, message_html=message_html
        )
        raw = self.gmail._encode_raw_message(msg)

        decoded = email.message_from_bytes(base64.urlsafe_b64decode(bytes(raw["raw"], "utf-8")))

        expected_items = [
            ("Content-Type", "multipart/alternative;\n boundary="),
            ("MIME-Version", "1.0"),
            ("to", to),
            ("from", sender),
            ("subject", subject),
        ]

        # The boundary id changes everytime. Replace it with the beginnig to
        # avoid failures
        updated_items = []
        for i in decoded.items():
            if "Content-Type" in i[0] and "multipart/alternative;\n boundary=" in i[1]:
                updated_items.append(("Content-Type", "multipart/alternative;\n boundary="))
            else:
                updated_items.append((i[0], i[1]))

        # Check the metadata
        self.assertListEqual(updated_items, expected_items)

        # Check the message
        # The first part is just a container for the text and html parts
        parts = decoded.get_payload()

        self.assertEqual(parts[0].get_payload(), message_text)
        self.assertEqual(parts[1].get_payload(), message_html)

        with open(f"{_dir}/assets/loremipsum_b64_jpeg.txt", "r") as f:
            b64_txt = f.read()
        self.assertEqual(parts[2].get_payload(), b64_txt)

        expected_id = f"<{attachments[0].split('/')[-1]}>"
        self.assertEqual(parts[2].get("Content-ID"), expected_id)
        self.assertEqual(parts[2].get_content_type(), "image/jpeg")

        # Check the number of parts
        expected_parts = 4
        self.assertEqual(sum([1 for i in decoded.walk()]), expected_parts)

    def test_create_message_attachments_m4a(self):
        sender = "Sender <sender@email.com>"
        to = "Recepient <recepient@email.com>"
        subject = "This is a test email with attachements"
        message_text = "The is the message text of the email with attachments"
        message_html = "<p>This is the html message part of the email with attachments</p>"
        attachments = [f"{_dir}/assets/loremipsum.m4a"]

        msg = self.gmail._create_message_attachments(
            sender, to, subject, message_text, attachments, message_html=message_html
        )
        raw = self.gmail._encode_raw_message(msg)

        decoded = email.message_from_bytes(base64.urlsafe_b64decode(bytes(raw["raw"], "utf-8")))

        expected_items = [
            ("Content-Type", "multipart/alternative;\n boundary="),
            ("MIME-Version", "1.0"),
            ("to", to),
            ("from", sender),
            ("subject", subject),
        ]

        # The boundary id changes everytime. Replace it with the beginnig to
        # avoid failures
        updated_items = []
        for i in decoded.items():
            if "Content-Type" in i[0] and "multipart/alternative;\n boundary=" in i[1]:
                updated_items.append(("Content-Type", "multipart/alternative;\n boundary="))
            else:
                updated_items.append((i[0], i[1]))

        # Check the metadata
        self.assertListEqual(updated_items, expected_items)

        # Check the message
        # The first part is just a container for the text and html parts
        parts = decoded.get_payload()

        self.assertEqual(parts[0].get_payload(), message_text)
        self.assertEqual(parts[1].get_payload(), message_html)

        with open(f"{_dir}/assets/loremipsum_b64_m4a.txt", "r") as f:
            b64_txt = f.read()
        self.assertEqual(parts[2].get_payload(), b64_txt)

        self.assertEqual(parts[2].get_content_maintype(), "audio")

        # Check the number of parts
        expected_parts = 4
        self.assertEqual(sum([1 for i in decoded.walk()]), expected_parts)

    def test_create_message_attachments_mp3(self):
        sender = "Sender <sender@email.com>"
        to = "Recepient <recepient@email.com>"
        subject = "This is a test email with attachements"
        message_text = "The is the message text of the email with attachments"
        message_html = "<p>This is the html message part of the email with attachments</p>"
        attachments = [f"{_dir}/assets/loremipsum.mp3"]

        msg = self.gmail._create_message_attachments(
            sender, to, subject, message_text, attachments, message_html=message_html
        )
        raw = self.gmail._encode_raw_message(msg)

        decoded = email.message_from_bytes(base64.urlsafe_b64decode(bytes(raw["raw"], "utf-8")))

        expected_items = [
            ("Content-Type", "multipart/alternative;\n boundary="),
            ("MIME-Version", "1.0"),
            ("to", to),
            ("from", sender),
            ("subject", subject),
        ]

        # The boundary id changes everytime. Replace it with the beginnig to
        # avoid failures
        updated_items = []
        for i in decoded.items():
            if "Content-Type" in i[0] and "multipart/alternative;\n boundary=" in i[1]:
                updated_items.append(("Content-Type", "multipart/alternative;\n boundary="))
            else:
                updated_items.append((i[0], i[1]))

        # Check the metadata
        self.assertListEqual(updated_items, expected_items)

        # Check the message
        # The first part is just a container for the text and html parts
        parts = decoded.get_payload()

        self.assertEqual(parts[0].get_payload(), message_text)
        self.assertEqual(parts[1].get_payload(), message_html)

        with open(f"{_dir}/assets/loremipsum_b64_mp3.txt", "r") as f:
            b64_txt = f.read()
        self.assertEqual(parts[2].get_payload(), b64_txt)

        self.assertEqual(parts[2].get_content_type(), "audio/mpeg")

        # Check the number of parts
        expected_parts = 4
        self.assertEqual(sum([1 for i in decoded.walk()]), expected_parts)

    def test_create_message_attachments_mp4(self):
        sender = "Sender <sender@email.com>"
        to = "Recepient <recepient@email.com>"
        subject = "This is a test email with attachements"
        message_text = "The is the message text of the email with attachments"
        message_html = "<p>This is the html message part of the email with attachments</p>"
        attachments = [f"{_dir}/assets/loremipsum.mp4"]

        msg = self.gmail._create_message_attachments(
            sender, to, subject, message_text, attachments, message_html=message_html
        )
        raw = self.gmail._encode_raw_message(msg)

        decoded = email.message_from_bytes(base64.urlsafe_b64decode(bytes(raw["raw"], "utf-8")))

        expected_items = [
            ("Content-Type", "multipart/alternative;\n boundary="),
            ("MIME-Version", "1.0"),
            ("to", to),
            ("from", sender),
            ("subject", subject),
        ]

        # The boundary id changes everytime. Replace it with the beginnig to
        # avoid failures
        updated_items = []
        for i in decoded.items():
            if "Content-Type" in i[0] and "multipart/alternative;\n boundary=" in i[1]:
                updated_items.append(("Content-Type", "multipart/alternative;\n boundary="))
            else:
                updated_items.append((i[0], i[1]))

        # Check the metadata
        self.assertListEqual(updated_items, expected_items)

        # Check the message
        # The first part is just a container for the text and html parts
        parts = decoded.get_payload()

        self.assertEqual(parts[0].get_payload(), message_text)
        self.assertEqual(parts[1].get_payload(), message_html)

        with open(f"{_dir}/assets/loremipsum_b64_mp4.txt", "r") as f:
            b64_txt = f.read()
        self.assertEqual(parts[2].get_payload(), b64_txt)

        self.assertEqual(parts[2].get_content_type(), "video/mp4")

        # Check the number of parts
        expected_parts = 4
        self.assertEqual(sum([1 for i in decoded.walk()]), expected_parts)

    def test_create_message_attachments_pdf(self):
        sender = "Sender <sender@email.com>"
        to = "Recepient <recepient@email.com>"
        subject = "This is a test email with attachements"
        message_text = "The is the message text of the email with attachments"
        message_html = "<p>This is the html message part of the email with attachments</p>"
        attachments = [f"{_dir}/assets/loremipsum.pdf"]

        msg = self.gmail._create_message_attachments(
            sender, to, subject, message_text, attachments, message_html=message_html
        )

        raw = self.gmail._encode_raw_message(msg)

        decoded = email.message_from_bytes(base64.urlsafe_b64decode(bytes(raw["raw"], "utf-8")))

        expected_items = [
            ("Content-Type", "multipart/alternative;\n boundary="),
            ("MIME-Version", "1.0"),
            ("to", to),
            ("from", sender),
            ("subject", subject),
        ]

        # The boundary id changes everytime. Replace it with the beginnig to
        # avoid failures
        updated_items = []
        for i in decoded.items():
            if "Content-Type" in i[0] and "multipart/alternative;\n boundary=" in i[1]:
                updated_items.append(("Content-Type", "multipart/alternative;\n boundary="))
            else:
                updated_items.append((i[0], i[1]))

        # Check the metadata
        self.assertListEqual(updated_items, expected_items)

        # Check the message
        # The first part is just a container for the text and html parts
        parts = decoded.get_payload()

        self.assertEqual(parts[0].get_payload(), message_text)
        self.assertEqual(parts[1].get_payload(), message_html)

        with open(f"{_dir}/assets/loremipsum_b64_pdf.txt", "r") as f:
            b64_txt = f.read()
        self.assertEqual(parts[2].get_payload(), b64_txt)

        self.assertEqual(parts[2].get_content_type(), "application/pdf")

        # Check the number of parts
        expected_parts = 4
        self.assertEqual(sum([1 for i in decoded.walk()]), expected_parts)

    def test__validate_email_string(self):
        emails = [
            {"email": "Sender <sender@email.com>", "expected": True},
            {"email": "sender@email.com", "expected": True},
            {"email": "<sender@email.com>", "expected": True},
            {"email": "Sender sender@email.com", "expected": False},
            {"email": "Sender <sender2email.com>", "expected": False},
        ]

        # The behavior of email.parseaddr depends on the python patch version
        # See https://github.com/python/cpython/issues/102988
        # or associated changelogs, e.g.
        # https://docs.python.org/3.8/whatsnew/changelog.html#python-3-8-20-final
        if getattr(email.utils, "supports_strict_parsing", False):
            emails.extend(
                [
                    {"email": "Sender <sender@email,com>", "expected": False},
                    {"email": "Sender <sender+alias@email,com>", "expected": False},
                ]
            )
        else:
            emails.extend(
                [
                    {"email": "Sender <sender@email,com>", "expected": True},
                    {"email": "Sender <sender+alias@email,com>", "expected": True},
                ]
            )

        for e in emails:
            if e["expected"]:
                self.assertTrue(self.gmail._validate_email_string(e["email"]))
            else:
                self.assertRaises(ValueError, self.gmail._validate_email_string, e["email"])

    # TODO test sending emails


if __name__ == "__main__":
    unittest.main()



import logging

import xmltodict
from bs4 import BeautifulSoup
from requests import HTTPError

from parsons import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector
from parsons.utilities.datetime import parse_date

logger = logging.getLogger(__name__)

MC_URI = "https://secure.mcommons.com/api/"
DATE_FMT = "%Y-%m-%d"


def _format_date(user_entered_date):
    if user_entered_date:
        formatted_date = parse_date(user_entered_date).strftime(DATE_FMT)
    else:
        formatted_date = None
    return formatted_date


class MobileCommons:
    """
    Instantiate the MobileCommons class.

    `Args:`
        username: str
            A valid email address connected to a  MobileCommons account. Not required if
            ``MOBILECOMMONS_USERNAME`` env variable is set.
        password: str
            Password associated with Zoom account. Not required if ``MOBILECOMMONS_PASSWORD``
            env variable set.
        company_id: str
            The company id of the MobileCommons organization to connect to. Not required if
            username and password are for an account associated with only one MobileCommons
            organization.
    """

    def __init__(self, username=None, password=None, company_id=None):
        self.username = check_env.check("MOBILECOMMONS_USERNAME", username)
        self.password = check_env.check("MOBILECOMMONS_PASSWORD", password)
        self.default_params = {"company": company_id} if company_id else {}
        self.client = APIConnector(uri=MC_URI, auth=(self.username, self.password))

    def _mc_get_request(
        self,
        endpoint,
        first_data_key,
        second_data_key,
        params,
        elements_to_unpack=None,
        limit=None,
    ):
        """
        A function for GET requests that handles MobileCommons xml responses and pagination

        `Args:`
            endpoint: str
                The endpoint, which will be appended to the base URL for each request
            first_data_key: str
                The first key used to extract the desired data from the response dictionary derived
                from the xml response. E.g. 'broadcasts'
            second_data_key: str
                The second key used to extract the desired data from the response dictionary derived
                from the xml response. The value of this key is a list of values. E.g. 'broadcast'
            params: dict
                Parameters to be passed into GET request
            elements_to_unpack: list
                A list of elements that contain dictionaries to be unpacked into new columns in the
                final table
            limit: int
                The maximum number of rows to return
        `Returns:`
            Parsons table with requested data
        """

        # Create a table to compile results from different pages in
        final_table = Table()
        # Max page_limit is 1000 for MC
        page_limit = min((limit or 1000), 1000)

        # Set get request params
        params = {"limit": page_limit, **self.default_params, **params}

        logger.info(f"Working on fetching first {page_limit} rows. This can take a long time.")

        # Make get call and parse XML into list of dicts
        page = 1
        response_dict = self._parse_get_request(endpoint=endpoint, params=params)

        # If there's only one row, then it is returned as a dict, otherwise as a list
        data = response_dict["response"][first_data_key][second_data_key]
        if isinstance(data, dict):
            data = [data]

        # Convert to parsons table
        response_table = Table(data)

        # empty_page used for pagination below
        if response_table.num_rows > 0:
            empty_page = False
        else:
            raise ValueError("There are no records for specified resource")

        # Unpack any specified elements
        if elements_to_unpack:
            for col in elements_to_unpack:
                response_table.unpack_dict(col)

        # Append to final table
        final_table.concat(response_table)
        final_table.materialize()

        # Now must paginate to get more records
        # MC GET responses sometimes includes a page_count parameter to indicate how many pages
        # are available. Other times, the response includes a 'num' parameter that, when
        # you reach an empty page, equals 0. In the first scenario we know to stop paginating when
        # we reach the last page. In the second, we know to stop paginating when we encounter
        # and empty page

        # First scenario
        try:
            avail_pages = int(response_dict["response"][first_data_key]["page_count"])
            total_records = avail_pages * page_limit
            page_indicator = "page_count"

        # Second scenario
        except KeyError:
            response_dict["response"][first_data_key]["num"]
            page_indicator = "num"
            # If page_count is not available, we cannot calculate total_records and will paginate
            # until we hit user defined limit or an empty page
            total_records = float("inf")

        # Go fetch other pages of data
        while final_table.num_rows < (limit or total_records) and not empty_page:
            page += 1
            page_params = {"page": str(page), **params}
            logger.info(
                f"Fetching rows {(page - 1) * page_limit + 1} - {(page) * page_limit} of {limit}"
            )
            # Send get request
            response_dict = self._parse_get_request(endpoint=endpoint, params=page_params)
            # Check to see if page was empty if num parameter is available
            if page_indicator == "num":
                empty_page = int(response_dict["response"][first_data_key]["num"]) == 0

            if not empty_page:
                # Extract data
                response_data = response_dict["response"][first_data_key][second_data_key]
                # When only one row of data it is returned as dict instead of list, which
                # cannot be put into table
                if isinstance(response_data, dict):
                    response_data = [response_data]

                response_table = Table(response_data)

                # Append to final table
                final_table.concat(response_table)
                final_table.materialize()

        return Table(final_table[:limit])

    def _check_response_status_code(self, response):
        """
        A helper function that checks the status code of a response and raises an error if the
        response code is not 200

        `Args:`
            response: requests package response object
        """
        if response.status_code != 200:
            error = f"Response Code {str(response.status_code)}"
            error_html = BeautifulSoup(response.text, features="html.parser")
            error += "\n" + error_html.h4.next
            error += "\n" + error_html.p.next
            raise HTTPError(error)

    def _parse_get_request(self, endpoint, params):
        """
        A helper function that sends a get request to MobileCommons and then parses XML responses in
        order to return the response as a dictionary

        `Args:`
            endpoint: str
                The endpoint, which will be appended to the base URL for each request
            params: dict
                Parameters to be passed into GET request
        `Returns:`
            xml response parsed into list or dictionary
        """
        response = self.client.request(endpoint, "GET", params=params)

        # If there's an error with initial response, raise error
        self._check_response_status_code(response)

        # If good response, compile data into final_table
        # Parse xml to nested dictionary and load to parsons table
        response_dict = xmltodict.parse(
            response.text, attr_prefix="", cdata_key="", dict_constructor=dict
        )
        return response_dict

    def _mc_post_request(self, endpoint, params):
        """
        A function for POST requests that handles MobileCommons xml responses

        `Args:`
            endpoint: str
                The endpoint, which will be appended to the base URL for each request
            params: dict
                Parameters to be passed into GET request
        `Returns:`
            xml response parsed into list or dictionary
        """

        response = self.client.request(endpoint, "POST", params=params)

        response_dict = xmltodict.parse(
            response.text, attr_prefix="", cdata_key="", dict_constructor=dict
        )
        if response_dict["response"]["success"] == "true":
            return response_dict["response"]
        else:
            raise HTTPError(response_dict["response"]["error"])

    def get_broadcasts(
        self, first_date=None, last_date=None, status=None, campaign_id=None, limit=None
    ):
        """
        A function for get broadcasts

        `Args:`
            first_date: str
                The date of the earliest possible broadcast you'd like returned. All common date
                format should work (e.g. mm/dd/yy or yyyy-mm-dd)
            last_date: str
                The date of the latest possible broadcast you'd like returned. All common date
                format should work (e.g. mm/dd/yy or yyyy-mm-dd)
            status: str
                'draft', 'scheduled', or 'generated'
            campaign_id: int
                Specify to return broadcasts from a specific campaign
            limit: int
                Max rows you want returned

        `Returns:`
            Parsons table with requested broadcasts
        """

        params = {
            "start_time": _format_date(first_date),
            "end_time": _format_date(last_date),
            "campaign_id": campaign_id,
            "status": status,
            **self.default_params,
        }

        return self._mc_get_request(
            endpoint="broadcasts",
            first_data_key="broadcasts",
            second_data_key="broadcast",
            params=params,
            elements_to_unpack=["campaign"],
            limit=limit,
        )

    def get_campaign_subscribers(
        self,
        campaign_id: int,
        first_date: str = None,
        last_date: str = None,
        opt_in_path_id: int = None,
        limit: int = None,
    ):
        """
        A function for getting subscribers of a specified campaign

        `Args:`
            campaign_id: int
                The campaign for which you'd like to get subscribers. You can get this from the url
                of the campaign's page after select a campaign at
                https://secure.mcommons.com/campaigns
            first_date: str
                The date of the earliest possible subscription returned. All common date
                format should work (e.g. mm/dd/yy or yyyy-mm-dd)
            last_date: str
                The date of the latest possible subscription you'd like returned. All common date
                format should work (e.g. mm/dd/yy or yyyy-mm-dd)
            opt_in_path_id: int
                Optional parameter to narrow results to on particular opt-in path. You can get this
                from the url of the opt in paths page https://secure.mcommons.com/opt_in_paths
            limit: int
                Max rows you want returned

        `Returns:`
            Parsons table with requested broadcasts
        """

        params = {
            "campaign_id": campaign_id,
            "from": _format_date(first_date),
            "to": _format_date(last_date),
            "opt_in_path_id": opt_in_path_id,
            **self.default_params,
        }

        return self._mc_get_request(
            endpoint="campaign_subscribers",
            first_data_key="subscriptions",
            second_data_key="sub",
            params=params,
            limit=limit,
        )

    def get_profiles(
        self,
        phones: list = None,
        first_date: str = None,
        last_date: str = None,
        include_custom_columns: bool = False,
        include_subscriptions: bool = False,
        limit: int = None,
    ):
        """
        A function for getting profiles, which are MobileCommons people records

        `Args:`
            phones: list
                A list of phone numbers including country codes for which you want profiles returned
                MobileCommons claims to recognize most formats.
            first_date: str
                The date of the earliest possible subscription returned. All common date
                format should work (e.g. mm/dd/yy or yyyy-mm-dd).
            last_date: str
                The date of the latest possible subscription you'd like returned. All common date
                format should work (e.g. mm/dd/yy or yyyy-mm-dd).
            include_custom_columns: bool
                Optional parameter to that, if set to True, will return custom column values for
                profiles as a list of dictionaries contained within a column.
            include_subscriptions: bool
                Optional parameter to that, if set to True, will return a list of campaigns a
                given profile is subscribed to in a single column
            limit: int
                Max rows you want returned

        `Returns:`
            Parsons table with requested broadcasts
        """

        custom_cols = "true" if include_custom_columns else "false"
        subscriptions = "true" if include_subscriptions else "false"

        if phones:
            phones = ",".join(phones)

        params = {
            "phone_number": phones,
            "from": _format_date(first_date),
            "to": _format_date(last_date),
            "include_custom_columns": custom_cols,
            "include_subscriptions": subscriptions,
            **self.default_params,
        }

        return self._mc_get_request(
            endpoint="profiles",
            first_data_key="profiles",
            second_data_key="profile",
            elements_to_unpack=["source", "address"],
            params=params,
            limit=limit,
        )

    def create_profile(
        self,
        phone,
        first_name=None,
        last_name=None,
        zip=None,
        addressline1=None,
        addressline2=None,
        city=None,
        state=None,
        opt_in_path_id=None,
        custom_column_values=None,
    ):
        """
        A function for creating or updating a single MobileCommons profile

        `Args:`
            phone: str
                Phone number to assign profile
            first_name: str
                Profile first name
            last_name: str
                Profile last name
            zip: str
                Profile 5-digit postal code
            addressline1: str
                Profile address line 1
            addressline2: str
                Profile address line 2
            city: str
                Profile city
            state: str
                Profile state
            opt_in_path_id: str
                ID of the opt-in path to send new profile through. This will determine the welcome
                text they receive.
            custom_column_values: dict
                Dictionary with custom column names as keys and custom column values
                as dictionary values

        `Returns:`
            ID of created/updated  profile
        """

        params = {
            "phone_number": phone,
            "first_name": first_name,
            "last_name": last_name,
            "postal_code": zip,
            "street1": addressline1,
            "street2": addressline2,
            "city": city,
            "state": state,
            "opt_in_path_id": opt_in_path_id,
            **self.default_params,
        }

        if custom_column_values:
            params.update(custom_column_values)

        response = self._mc_post_request("profile_update", params=params)
        return response["profile"]["id"]



import logging
import time

from parsons.etl import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

logger = logging.getLogger(__name__)

POLLING_DELAY = 1
ACTBLUE_API_ENDPOINT = "https://secure.actblue.com/api/v1"


class ActBlue(object):
    """
    Instantiate class.

       `Args:`
            actblue_client_uuid: str
                The ActBlue provided Client UUID. Not required if ``ACTBLUE_CLIENT_UUID`` env
                variable set.
            actblue_client_secret: str
                The ActBlue provided Client Secret. Not required if ``ACTBLUE_CLIENT_SECRET`` env
                variable set.
            actblue_uri: str
                The URI to access the CSV API. Not required, default is
                https://secure.actblue.com/api/v1. You can set an ``ACTBLUE_URI`` env variable or
                use this URI parameter if a different endpoint is necessary - for example, when
                running this code in a test environment where you don't want to hit the actual API.
            max_retries: int
                The maximum number of times to poll the API for a download URL. Not required, default
                is None, which means it will poll indefinitely until a download URL is returned.
                ```ACTBLUE_MAX_RETRIES``` env variable can be set, which will override this parameter.
        For instructions on how to generate a Client UUID and Client Secret set,
        visit https://secure.actblue.com/docs/csv_api#authentication.
    """

    def __init__(
        self,
        actblue_client_uuid=None,
        actblue_client_secret=None,
        actblue_uri=None,
        max_retries=None,
    ):
        self.actblue_client_uuid = check_env.check("ACTBLUE_CLIENT_UUID", actblue_client_uuid)
        self.actblue_client_secret = check_env.check("ACTBLUE_CLIENT_SECRET", actblue_client_secret)
        self.uri = (
            check_env.check("ACTBLUE_URI", actblue_uri, optional=True) or ACTBLUE_API_ENDPOINT
        )
        self.headers = {
            "accept": "application/json",
        }
        self.client = APIConnector(
            self.uri,
            auth=(self.actblue_client_uuid, self.actblue_client_secret),
            headers=self.headers,
        )
        self.max_retries = check_env.check("ACTBLUE_MAX_RETRIES", max_retries, optional=True)
        self.max_retries = int(self.max_retries) if self.max_retries else None

    def post_request(self, csv_type=None, date_range_start=None, date_range_end=None):
        """
        POST request to ActBlue API to begin generating the CSV.

        `Args:`
            csv_type: str
                Type of CSV you are requesting.
                Options:
                    'paid_contributions': contains paid, non-refunded contributions to the entity
                    (campaign or organization) you created the credential for, during the specified
                    date range.

                    'refunded_contributions': contributions to your entity that were refunded,
                    during the specified date range.

                    'managed_form_contributions': contributions made through any form that is
                    managed by your entity, during the specified date range - including
                    contributions to other entities via that form if it is a tandem form.
            date_range_start: str
                Start of date range to withdraw contribution data (inclusive). Ex: '2020-01-01'
            date_range_end: str
                End of date range to withdraw contribution data (exclusive). Ex: '2020-02-01'

        `Returns:`
            Response of POST request; a successful response includes 'id', a unique identifier for
            the CSV being generated.
        """

        body = {
            "csv_type": csv_type,
            "date_range_start": date_range_start,
            "date_range_end": date_range_end,
        }
        logger.info(f"Requesting {csv_type} from {date_range_start} up to {date_range_end}.")
        response = self.client.post_request(url="csvs", json=body)
        return response

    def get_download_url(self, csv_id=None):
        """
        GET request to retrieve download_url for generated CSV.

        `Args:`
            csv_id: str
                Unique identifier of the CSV you requested.

        `Returns:`
            While CSV is being generated, 'None' is returned. When CSV is ready, the method returns
            the download_url.
        """
        response = self.client.get_request(url=f"csvs/{csv_id}")
        if response.get("download_url") is None and response.get("status") != "in_progress":
            raise ValueError("CSV generation failed: %s", response)

        return response["download_url"]

    def poll_for_download_url(self, csv_id):
        """
        Poll the GET request method to check whether CSV generation has finished, signified by the
        presence of a download_url.

        `Args:`
            csv_id: str
                Unique identifier of the CSV you requested.

        `Returns:`
            Download URL from which you can download the generated CSV, valid for 10 minutes after
            retrieval. Null until CSV has finished generating. Keep this URL secure because until
            it expires, it could be used by anyone to download the CSV.
        """

        logger.info("Request received. Please wait while ActBlue generates this data.")
        download_url = None
        tries = 0
        while download_url is None and (self.max_retries is None or tries < self.max_retries):
            download_url = self.get_download_url(csv_id)
            time.sleep(POLLING_DELAY)
            tries += 1

        if download_url is None:
            raise TimeoutError("CSV generation timed out. Increase max_retries and try again.")

        logger.info("Completed data generation.")
        logger.info("Beginning conversion to Parsons Table.")
        return download_url

    def get_contributions(self, csv_type, date_range_start, date_range_end, **csvargs):
        """
        Get specified contribution data from CSV API as Parsons table.

        `Args:`
            csv_type: str
                Type of CSV you are requesting.
                Options:
                    'paid_contributions': contains paid, non-refunded contributions to the entity
                    (campaign or organization) you created the credential for, during the specified
                    date range.

                    'refunded_contributions': contributions to your entity that were refunded,
                    during the specified date range.

                    'managed_form_contributions': contributions made through any form that is
                    managed by your entity, during the specified date range - including
                    contributions to other entities via that form if it is a tandem form.
            date_range_start: str
                Start of date range to withdraw contribution data (inclusive). Ex: '2020-01-01'
            date_range_end: str
                End of date range to withdraw contribution data (exclusive). Ex: '2020-02-01'
            **csvargs:
                Any additional arguments will be passed to Table.from_csv as keyword arguments.

        `Returns:`
            Contents of the generated contribution CSV as a Parsons table.
        """

        post_request_response = self.post_request(csv_type, date_range_start, date_range_end)
        csv_id = post_request_response["id"]
        download_url = self.poll_for_download_url(csv_id)
        table = Table.from_csv(download_url, **csvargs)
        logger.info("Completed conversion to Parsons Table.")
        return table



import logging

from parsons import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

logger = logging.getLogger(__name__)


class Census(object):
    """
    Class that creates a connector to the Census Bureau API
    """

    def __init__(self, api_key=None):
        """
        Instantiate Census class.

           Args:
                api_key: string, key for Census API access
                 (optional, can also be pulled from environment variable CENSUS_API_KEY)
        """
        self.api_key = check_env.check("CENSUS_API_KEY", api_key)
        self.host = "https://api.census.gov/data"

    def get_census(self, year=None, dataset_acronym=None, variables=None, location=None):
        """
        Pull census data. For background check out the `Census API Guide
                <https://www.census.gov/data/developers/guidance/api-user-guide.html>`_

        Args:
                year: 4-digit string or integer
                        e.g. '2019' or 2019

                dataset_acronym: string with dataset name
                        e.g. '/acs/acs1'

                variables: comma-separated string with variable names
                        e.g. 'NAME,B01001_001E'

                location: string with desired locations
                        e.g. 'us:*'

        Return:
                Parsons table with data
        """
        # set up the URL
        g = "?get="
        usr_key = f"&key={self.api_key}"
        year = str(year)  # in case someone passes int
        location = "&for=" + location
        query_url = f"{self.host}/{year}{dataset_acronym}{g}{variables}{location}{usr_key}"

        # create connector
        connector = APIConnector(uri=self.host)

        # pull data
        response = connector.get_request(url=query_url)

        # turn into Parsons table and return it
        return Table(response)
